"""Implements plot functions for agent-based models, e.g. of agents and their
movement in a domain."""

import copy
import itertools
import logging
from collections import deque
from math import sqrt
from typing import Dict, Optional, Sequence, Tuple, Union

import numpy as np
import xarray as xr
from dantro.base import BaseDataContainer
from dantro.plot import ColorManager
from dantro.tools import recursive_update
from matplotlib.collections import LineCollection, PathCollection
from matplotlib.legend import Legend
from matplotlib.legend_handler import HandlerBase, HandlerPathCollection
from matplotlib.markers import MarkerStyle
from matplotlib.patches import Patch
from matplotlib.path import Path
from matplotlib.transforms import AffineDeltaTransform

from ...tools import ensure_dict, is_iterable
from .. import PlotHelper, is_plot_func
from ._mpl import adjust_figsize_to_aspect

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

MARKERS: Dict[str, Path] = {
    "wedge": Path(
        [(0.0, -0.1), (0.3, 0.0), (0.0, 0.1), (0.0, 0.0)],
        codes=[Path.MOVETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY],
    ),
    "fish": Path(
        [
            (-5, 0),
            (-2, +0.2),
            (+3, +2),
            (+3, 0),
            #
            (+3, -2),
            (-2, -0.2),
            (-5, 0),
        ],
        codes=[
            Path.MOVETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            #
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
        ],
    ),
    "fish2": Path(
        [
            (-5, 0),
            (-2, +0.2),
            (+3, +2.5),
            (+3, 0),
            #
            (+3, -2.5),
            (-2, -0.2),
            (-5, 0),
            #
            # tail
            (-3, 0),
            (-5.5, -1.5),
            (-5, 0),
            (-5.5, +1.5),
            #
            # fins left & right
            (+2, 0),
            (-2, -1.3),
            (-1, 0),
            (-2, +1.3),
        ],
        codes=[
            Path.MOVETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            #
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
            #
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
            #
            Path.MOVETO,
            Path.LINETO,
            Path.LINETO,
            Path.LINETO,
        ],
    ),
}
"""Custom markers that can be used in :py:class:`.AgentCollection`, an
essential part of :py:func:`.abmplot`.

Available markers:

- ``wedge``: A triangle that can be used to denote orientations.
- ``fish``: A simple fish-like shape.
- ``fish2``: A more elaborate fish-like shape.

.. image:: ../_static/_gen/abmplot/markers.pdf
    :target: ../_static/_gen/abmplot/markers.pdf
    :width: 100%

.. hint::

    To have consistent orientations, note that markers should point along the
    positive x-axis. With increasing orientation values, they are rotated in
    counter-clockwise direction.
    Angles are measured in radians.

.. note::

    Suggestions for additional markers are
    `welcome <https://gitlab.com/utopia-project/utopya/-/issues>`_.
"""
# TODO add ant-like shape
# TODO tweak shapes to consist only of a single path, not two overlapping ones


class AgentCollection(PathCollection):
    """A collection of agent markers and (optionally) their position history
    in the form of a tail.
    Combines :py:class:`~matplotlib.collections.PathCollection` with a
    :py:class:`~matplotlib.collections.LineCollection` for the tails.
    """

    def __init__(
        self,
        xy: Union[np.ndarray, Sequence[Union[Path, Tuple[float, float]]]],
        *,
        marker: Union[Path, Patch, str, dict] = "auto",
        default_marker: Union[Path, Patch, str, dict] = "o",
        default_marker_oriented: Union[Path, Patch, str, dict] = "wedge",
        sizes: Union[float, Sequence[float]] = None,
        size_scale: float = 1.0,
        normalize_marker_sizes: bool = True,
        orientations: Union[float, Sequence[float]] = None,
        base_orientation: float = 0.0,
        tail_length: int = 0,
        tail_kwargs: dict = None,
        tail_decay: float = 0.0,
        tail_max_segment_length: float = None,
        use_separate_paths: bool = True,
        **kwargs,
    ):
        """Set up an AgentCollection, visualizing agents and (optionally) their
        position history using a tail.

        .. hint::

            If you need a higher performance for this plot, consider not using
            the *decaying* tails, for which the line segments need to be drawn
            individually.

        Args:
            xy (Sequence[Tuple[float, float]]): The positions of the agents.
                The number of agents is extracted from the length of this
                sequence or ``(x, y)`` positions or ``(N, 2)`` array.
            marker (Union[matplotlib.path.Path, matplotlib.patches.Patch, str, dict], optional):
                The marker to use to denote agent positions. The following
                input types are possible:

                - A :py:class:`~matplotlib.path.Path` is used as it is
                - A :py:class:`dict` is used as keyword arguments to the
                  :py:class:`~matplotlib.path.Path`, where ``vertices`` can be
                  a sequence of 2-tuples that defines the vertices of the path.
                - A :py:class:`~matplotlib.patches.Patch` already defines a
                  path, which is extracted and used as marker. All other
                  properties are ignored.
                - :py:class:`str` is looked up in :py:data:`.MARKERS`; if no
                  such entry is available, the remaining options are evaluated.
                - All other input is used to construct a
                  :py:class:`~matplotlib.markers.MarkerStyle` from which the
                  path is extracted.

            default_marker (Union[matplotlib.path.Path, matplotlib.patches.Patch, str, dict], optional):
                The marker to use in case that ``marker`` is None and agents
                are *not* initialized with an orientation.
            default_marker_oriented (Union[matplotlib.path.Path, matplotlib.patches.Patch, str, dict], optional):
                The marker to use in case that ``marker`` is None and agents
                *are* initialized with an orientation.
            sizes (Union[float, Sequence[float]], optional): Agent sizes in
                units of area (of bounding box) squared.
                If a scalar value is given, all agents will have the same size.
            size_scale (float, optional): A scaling factor for the sizes.
            normalize_marker_sizes (bool, optional): If True, will normalize
                the size of the given ``marker``. This is done by computing
                the area of the bounding box and scaling the path such that the
                bounding box has an area of one.
            orientations (Union[float, Sequence[float]], optional): Single or
                individual orientations of the agents. Zero corresponds to
                pointing along the positive x-axis, with increasing orientation
                values leading to a counter-clockwise rotation. Angles need to
                be given in radians.
            base_orientation (float, optional): Rotates the given marker by
                this value such that all ``orientations`` have a constant
                offset.
            tail_length (int, optional): For non-zero integers, will draw a
                tail behind the agents that traces their position history.
                This is only relevant if the collection is used as part of an
                animation where the position history can be aggregated upon
                each call to :py:meth:`.set_offsets`.
            tail_kwargs (dict, optional): Keyword arguments that are passed to
                the :py:class:`~matplotlib.collections.LineCollection` which is
                used to represent the tails.
            tail_decay (float, optional): A decay factor. If non-zero, will
                multiply each consecutive tail segments ``alpha`` value with
                ``(1 - tail_decay)``.
            tail_max_segment_length (float, optional): If set, will *not* draw
                tail segments that are longer than this value in either the
                x- or the y-direction. This can be useful if agents move in a
                periodic space where positions may jump.
            use_separate_paths (bool, optional): If False, only a single path
                is used to represent all agents in the underlying
                :py:class:`~matplotlib.collections.PathCollection`.

                .. warning::
                    Setting this to False should *only* be done if all agents
                    can be thought of as identical, i.e. have the same size,
                    color, marker, and orientation — otherwise this leads to
                    subtle and seemingly random breakage in the form of
                    markers not being drawn.
            **kwargs: Passed on to
                :py:class:`~matplotlib.collections.PathCollection`
        """
        xy = np.array(xy)
        if xy.ndim != 2 or xy.shape[1] != 2:
            raise ValueError(
                "Agent positions need to be of shape `(N, 2)` but were "
                f"{xy.shape}! Given data was:\n{str(xy)}"
            )

        # Attributes
        self._normalize_marker_sizes = normalize_marker_sizes
        self._size_scale = size_scale
        self._size_factors = None

        self._has_orientation = orientations is not None
        self._R = np.empty((2, 2))  # re-usable rotation matrix
        self._base_vertices = None
        self._base_orientation = base_orientation
        self._orientations = None

        self._tail_length = tail_length
        self._tails = None
        self._tail_kwargs = ensure_dict(tail_kwargs)
        self._tail_decay = tail_decay
        self._tail_max_segment_length = tail_max_segment_length
        self._offsets_history = None

        # Decide whether they need to be oriented, in which case we need many
        # separate Path objects. Otherwise, a single Path suffices.
        if marker == "auto":
            marker = (
                default_marker_oriented
                if self.has_orientation
                else default_marker
            )
        self._markerpath = self._prepare_marker_path(marker)
        self._has_markers = bool(marker)

        if use_separate_paths or self.has_orientation:
            paths = [copy.deepcopy(self.markerpath) for _ in range(len(xy))]

        else:
            # A bit simpler, only need a single marker
            # FIXME This causes issues in some scenarios, e.g. by mapping
            #       the color only to the first path or not drawing
            #       markers at all!
            #       Perhaps always use separate markerpaths?!
            paths = [self.markerpath]

        # Compute size factors to have reasonably sized markers (and not
        # markers in data units, which typically does not make a lot of sense)
        self._size_factors = self._compute_size_factors(paths)

        # Now setup the underlying PathCollection
        super().__init__(paths, offsets=xy, sizes=self._size_factors, **kwargs)

        # Optionally, set sizes
        if sizes is not None:
            self.set_markersizes(sizes)

        # Optionally, set orientations
        if self.has_orientation:
            # Store copies of the base vertices which will be used for
            # rotating the markers. These base vertices are used for the
            # rotation in order to not accumulate rounding errors etc.
            self._base_vertices = [p.vertices.copy() for p in self.get_paths()]

            # Base vertices should not be writeable, but the patches' vertices
            # need to be writeable. Take care of that here, reduces repetitions
            for _bv in self._base_vertices:
                _bv.setflags(write=False)  # for safety

            # Can apply orientations now
            self.set_orientations(orientations)

        # Optionally, prepare for drawing tails
        if self.tail_length:
            self._offsets_history = deque([], self.tail_length + 1)
            self._add_to_offsets_history(self.get_offsets())

            if not self._tail_decay:
                self._segments = np.empty((len(self), self.tail_length + 1, 2))
            else:
                self._segments = np.empty((len(self) * self.tail_length, 2, 2))
            self._segments.fill(np.nan)

            self.draw_tails()

    def __len__(self) -> int:
        """Number of agents associated with this AgentCollection"""
        return len(self.get_offsets())

    @property
    def has_orientation(self) -> bool:
        """Whether agent orientation is represented"""
        return self._has_orientation

    @property
    def has_markers(self) -> bool:
        """Whether a marker will be shown; if disabled, can skip setting array
        data and thus increase performance somewhat ..."""
        return self._has_markers

    @property
    def tail_length(self) -> int:
        """Length of the tail of each agent, i.e. the number of *historic*
        offset positions that are connected into a line."""
        return self._tail_length

    @property
    def tails(self) -> Optional[LineCollection]:
        """The :py:class:`~matplotlib.collections.LineCollection` that
        represents the tails of the agents — or None if tails are not
        activated.

        .. note::

            This collection still needs to be registered *once* with the axes
            via :py:meth:`~matplotlib.axes.Axes.add_collection`.
        """
        return self._tails

    @property
    def markerpath(self) -> Optional[Path]:
        """Returns the used markerpath"""
        return self._markerpath

    def draw_tails(self, **kwargs) -> LineCollection:
        """Draws the tails *from scratch*, removing the currently added
        collection if there is one.

        .. note::

            If calling this, the collection representing the tails needs to be
            re-added to the desired axes. To update tail positions according
            to the offsets history, use :py:meth:`.update_tails`.
        """
        if not self.tail_length:
            raise ValueError(
                "Cannot add tails to AgentCollection after initialization!"
            )

        if kwargs:
            kwargs = recursive_update(copy.deepcopy(self._tail_kwargs), kwargs)
        else:
            kwargs = self._tail_kwargs

        if self._tails is not None:
            try:
                self._tails.remove()
            except:
                pass

        self._tails = LineCollection(self._build_line_segments(), **kwargs)

        if self._tail_decay != 0:
            alpha = self.tails.get_alpha()
            if not isinstance(alpha, float):
                alpha = 1.0
            decaying_alpha = itertools.accumulate(
                [alpha] * self.tail_length,
                lambda a, _: a * (1.0 - self._tail_decay),
            )
            self.tails.set_alpha(list(decaying_alpha) * len(self))

        return self.tails

    def update_tails(self) -> LineCollection:
        """Updates the collection that holds the tails using the offsets
        history.
        """
        if not self.tail_length:
            raise ValueError(
                "Cannot add tails to AgentCollection after initialization!"
            )
        self.tails.set_segments(self._build_line_segments())
        return self.tails

    def set_markersizes(self, sizes):
        """Sets the marker sizes, taking into account the size scaling.

        .. note::

            This behaves differently from the inherited ``set_sizes`` method
            and thus has a different name. The difference is that
            :py:meth:`~matplotlib.collections.PathCollection.set_sizes` is also
            called during saving of a plot and would thus lead to repeated
            application of the size scaling, which is not desired.
        """
        if not is_iterable(sizes):
            sizes = np.array([sizes] * len(self))

        super().set_sizes(self._size_factors * sizes)

    def set_offsets(self, offsets) -> None:
        """Sets the offsets, i.e. path positions of all agents.

        Invokes :py:meth:`~matplotlib.collections.PathCollection.set_offsets`
        and afterwards stores the offset history, using it for
        :py:meth:`.update_tails`.
        """
        super().set_offsets(offsets)

        if self.tail_length:
            self._add_to_offsets_history(offsets)
            self.update_tails()

    def set_orientations(
        self, orientations: Union[float, Sequence[float]]
    ) -> None:
        """Set the orientations of all agents. If a scalar is given, will set
        all orientations to that value. If any kind of sequence is given, will
        associate individual orientations with agents.

        .. note::

            This can only be used if the collection was *initialized* with
            ``orientations``, see :py:meth:`.__init__`.
        """
        if not self.has_orientation:
            raise ValueError(
                "AgentCollection was not initialized with orientations; "
                "cannot change them after initialization!"
            )

        if not is_iterable(orientations):
            orientations = [orientations] * len(self._base_vertices)

        paths = self.get_paths()

        if not (len(paths) == len(orientations) == len(self._base_vertices)):
            raise ValueError(
                f"There was a length mismatch between the number of paths in "
                f"this collection ({len(paths)}), the given values for the "
                f"new orientations ({len(orientations)}), and the number of "
                f"stored base vertices ({len(self._base_vertices)})!"
            )

        for path, _o, _bv in zip(paths, orientations, self._base_vertices):
            self._rotate_path(path, orientation=_o, base_vertices=_bv)

        self._orientations = orientations

    def get_orientations(self) -> Sequence[float]:
        """Returns the current orientations of all agents.

        .. note::

            Changing the returned object does not rotate the agents!
            Use :py:meth:`.set_orientations` to achieve that.
        """
        if not self.has_orientation:
            raise ValueError(
                "AgentCollection was not initialized with orientations!"
            )
        return self._orientations

    # .........................................................................

    def _prepare_marker_path(
        self, marker: Union[Path, Patch, str, list, dict]
    ) -> Path:
        """Generates the :py:class:`~matplotlib.path.Path` needed for the
        agent marker using a variety of setup methods, depending on argument
        type:

        - :py:class:`~matplotlib.path.Path` is used as it is
        - :py:class:`dict` is used to call :py:class:`~matplotlib.path.Path`
        - :py:class:`~matplotlib.patches.Patch` already defines a path, which
          is extracted and used.
        - :py:class:`str` is looked up in :py:data:`.MARKERS`; if no such entry
          is available, the remaining options are evaluated.
        - Other types construct a :py:class:`~matplotlib.markers.MarkerStyle`
          from which the path is extracted.

        After constructing the path, applies the base orientation by rotating
        it by that value.
        """
        if isinstance(marker, Path):
            path = marker

        elif isinstance(marker, dict):
            path = Path(**marker)

        elif isinstance(marker, Patch):
            path = copy.deepcopy(marker.get_path())

        elif isinstance(marker, str) and marker in MARKERS:
            path = copy.deepcopy(MARKERS[marker])

        else:
            path = MarkerStyle(marker).get_path()

        path.vertices.setflags(write=True)
        self._rotate_path(
            path,
            base_vertices=path.vertices.copy(),
            orientation=self._base_orientation,
        )
        return path

    def _compute_size_factors(self, paths: Sequence[Path]) -> np.ndarray:
        """Given a sequence of Paths, computes the corresponding size factors
        which are used in :py:meth:`.set_markersizes` to scale the individual
        marker sizes. The resulting sizes are multiplied with the given
        size scale.

        If marker size normalization is activated, size factors may vary
        between the markers.
        """

        def bbox_area(p) -> float:
            bbox = p.get_extents()
            return bbox.height * bbox.width

        if not self._normalize_marker_sizes:
            return self._size_scale * np.ones((len(paths),))

        return self._size_scale * np.array([1 / bbox_area(p) for p in paths])

    def _add_to_offsets_history(self, new_offsets: np.ndarray) -> None:
        """Adds the given offset to the left side of the deque, pushing old
        entries off the right side of the deque.

        If the tail segment length is limited, will mask the offset positions
        for individual agents if they differ by more than that value from
        their previous position.
        *Note* that in this case the history is no longer strictly correct; it
        is already evaluated for use in :py:meth:`.draw_tails` in order to
        increase performance and avoid evaluating this over and over again.
        """
        max_delta = self._tail_max_segment_length
        if max_delta is not None and len(self._offsets_history) >= 1:
            # Need to mask offset jumps larger than the maximum delta.
            # This will lead to tail segments not being drawn.
            new_offsets = new_offsets.copy()
            abs_delta = np.abs(self._offsets_history[0] - new_offsets)
            new_offsets[abs_delta > max_delta] = np.nan

        self._offsets_history.appendleft(new_offsets)

    def _build_line_segments(self) -> np.ndarray:
        """Creates the line segments for the tails using the offsets history"""
        TL = self._tail_length
        if not self._tail_decay:
            # Can simply use the offsets history (with changed slicing)
            for i, _offsets in enumerate(self._offsets_history):
                self._segments[:, TL - i, :] = _offsets

        else:
            # Have a more complicated segments structure:
            #   axis 0: [0, TL)    segments of first agent's tail,
            #           [TL, 2TL)  segments of second agent's tail,
            #           etc., with lower indices being for more recent history.
            #   axis 1: start and end point
            #   axis 2: x and y coordinates (of start and end point)
            hist = list(self._offsets_history)
            for i, (_p0, _p1) in enumerate(zip(hist[:-1], hist[1:])):
                # _p0 and _p1: offsets of *all* agents at adjacent times
                self._segments[i::TL, 0, :] = _p0
                self._segments[i::TL, 1, :] = _p1

        return self._segments

    def _rotate_path(
        self, path: Path, *, base_vertices: np.ndarray, orientation: float
    ) -> None:
        """Sets the given path's vertices by rotating (a copy of) the base
        vertices by the given ``orientation`` (in radians)."""
        _cos, _sin = np.cos(orientation), np.sin(orientation)
        self._R[0, 0] = _cos
        self._R[1, 0] = _sin
        self._R[0, 1] = -_sin
        self._R[1, 1] = _cos
        path._vertices[...] = np.dot(base_vertices, self._R.T)
        path._orientation = orientation


class HandlerAgentCollection(HandlerPathCollection):
    """Legend handler for :py:class:`.AgentCollection` instances."""

    def create_collection(
        self, orig_handle, sizes, offsets, offset_transform
    ) -> AgentCollection:
        """Returns an AgentCollection from which legend artists are created."""
        ac = AgentCollection(
            offsets,
            marker=orig_handle.markerpath,
            sizes=sizes,  # FIXME too small?
            offset_transform=offset_transform,
        )
        # TODO What about the colormap etc?
        return ac

    def create_artists(
        self,
        legend,
        orig_handle,
        xdescent,
        ydescent,
        width,
        height,
        fontsize,
        trans,
    ):
        """Creates the artists themselves.

        For original docstring see
        :py:meth:`matplotlib.legend_handler.HandlerBase.create_artists`."""
        xdata, xdata_marker = self.get_xdata(
            legend, xdescent, ydescent, width, height, fontsize
        )

        ydata = self.get_ydata(
            legend, xdescent, ydescent, width, height, fontsize
        )

        sizes = self.get_sizes(
            legend, orig_handle, xdescent, ydescent, width, height, fontsize
        )

        p = self.create_collection(
            orig_handle,
            sizes,
            offsets=list(zip(xdata_marker, ydata)),
            offset_transform=trans,
        )

        self.update_prop(p, orig_handle, legend)
        p.set_offset_transform(trans)
        return [p]


Legend.update_default_handler_map({AgentCollection: HandlerAgentCollection()})


# -----------------------------------------------------------------------------


def _get_data(d: Union[xr.DataArray, xr.Dataset], var: str) -> xr.DataArray:
    """Retrieve data by key, e.g. for hue, size, ..."""
    try:
        return d[var]

    except Exception as exc:
        if isinstance(d, xr.DataArray):
            if d.name == var:
                return d
            raise ValueError(
                "If using xr.DataArray for agent data, make sure to have the "
                f"array's `name` attribute set (in this case to: '{var}')!\n"
                f"Got:  {str(d)}"
            ) from exc

        raise ValueError(
            f"Failed retrieving agent data variable '{var}' from:\n{d}\n"
            f"Got a {type(exc).__name__}: {exc}\n"
            f"Make sure that the data variable '{var}' exists "
            f"or pass an xr.DataArray with name '{var}' instead."
        ) from exc


def _parse_vmin_vmax(
    specs: dict,
    d: Union[xr.DataArray, xr.Dataset],
    *,
    vmin_key: str = "vmin",
    vmax_key: str = "vmax",
) -> dict:
    """Adapts ``specs`` dict entries ``vmin_key`` and ``vmax_key``, replacing
    their *values* with the minimum or maximum value of given data ``d``."""
    if specs.get(vmin_key) == "min":
        specs[vmin_key] = d.min().item()
    if specs.get(vmax_key) == "max":
        specs[vmax_key] = d.max().item()

    return specs


def _set_domain(
    *,
    ax: "matplotlib.axes.Axes",
    layers: dict,
    mode: str = "auto",
    extent: Tuple[float, float, float, float] = None,
    height: float = None,
    aspect: Union[float, int, str] = "auto",
    update_only: bool = False,
    pad: float = 0.05,
) -> Optional[Tuple[float, float]]:
    """Sets the axis limits and computes the area of the domain

    Args:
        ax (matplotlib.axes.Axes): The axes to adjust
        layers (dict): Layer information prepared by :py:func:`.abmplot`
        mode (str, optional): Domain mode, can be ``auto``, ``manual``,
            ``fixed`` and ``follow``. In fixed mode, all available data in
            ``layers`` is inspected to derive domain bounds.
            In ``follow`` mode, the currently used collection positions are
            used to determine the boundaries.
            In ``auto`` mode, will use a ``fixed`` domain if *no* ``extent``
            was given, otherwise ``manual`` mode is used.
        extent (Tuple[float, float, float, float], optional): A manually set
            extent in form ``(left, right, bottom, top)``. Alternatively,
            a 2-tuple will be interpreted as ``(0, right, 0, top)``.
        height (float, optional): The height of the domain in data units
        aspect (Union[float, int, str], optional): The aspect ratio of the
            domain, the width being computed via ``height * aspect``.
            If ``auto``, will
        update_only (bool, optional): If true, will only make changes to the
            domain bounds if necessary, e.g. in ``follow`` mode.
        pad (float, optional): Relative padding added to each side.

    Returns:
        Optional[Tuple[float, float]]:
            The domain area (in data units squared) and the aspect ratio, i.e.
            ``width/height``. The return value will be None if ``update_only``
            was set.
    """
    # Determine mode
    if mode == "auto":
        mode = "fixed" if extent is None else "manual"

    # May already have a new extent specified
    if extent is not None:
        try:
            _l, _r, _b, _t = extent

        except:
            try:
                _l, _b = 0, 0
                _r, _t = extent
            except:
                raise

    else:
        # Get the "previous" values, just to have something set
        _l, _r = ax.get_xlim()
        _b, _t = ax.get_ylim()

    # Depending on mode, the values may further change
    if mode is None or mode == "manual":
        if update_only:
            return

    elif isinstance(mode, str):
        # Decide which data source to use
        colls = [lyr["coll"] for lyr in layers.values()]
        have_colls = all([c is not None for c in colls])

        if mode == "fixed" and update_only:
            # No need to update
            return

        if mode == "fixed" or (mode == "follow" and not have_colls):
            # Get all the data
            # NOTE Will need adaption if plotting on multiple axes
            xdata = [lyr["data"][lyr["x"]] for lyr in layers.values()]
            ydata = [lyr["data"][lyr["y"]] for lyr in layers.values()]

        elif mode == "follow":
            # Get the existing offset data from the collection
            xdata = [coll.get_offsets()[:, 0] for coll in colls]
            ydata = [coll.get_offsets()[:, 1] for coll in colls]

        else:
            raise ValueError(
                f"Got invalid `mode` argument with value '{mode}'! "
                "Possible modes:  fixed, follow, manual, auto"
            )

        # Determine bounds
        _l = min([x.min().item() for x in xdata])
        _r = max([x.max().item() for x in xdata])
        _b = min([y.min().item() for y in ydata])
        _t = max([y.max().item() for y in ydata])

    else:
        raise TypeError(
            f"Got invalid `mode` argument of type {type(mode)}, expected str!"
        )

    # Ensure the domain has a fixed aspect ratio
    if aspect == "auto":
        if mode == "follow":
            aspect = 1.0
        else:
            aspect = abs(_r - _l) / abs(_t - _b)

    if aspect is not None:
        current_width = abs(_r - _l)
        current_height = abs(_t - _b)

        if height:
            target_width = height * aspect
            target_height = height
        else:
            target_width = max(current_width, current_height * aspect)
            target_height = max(current_height, current_width / aspect)

        if current_width < target_width:
            d = target_width - current_width
            _l -= d / 2
            _r += d / 2

        if current_height < target_height:
            d = target_height - current_height
            _b -= d / 2
            _t += d / 2

    # Apply padding
    if pad is not None:
        _w = abs(_r - _l)
        _h = abs(_t - _b)
        _l -= _w * pad
        _r += _w * pad
        _b -= _h * pad
        _t += _h * pad

    # Finally, compute domain area and actually set the limits
    domain_area = abs(_r - _l) * abs(_t - _b)
    actual_aspect = abs(_r - _l) / abs(_t - _b)
    ax.set_xlim(_l, _r)
    ax.set_ylim(_b, _t)

    return domain_area, actual_aspect


# -----------------------------------------------------------------------------


def draw_agents(
    d: Union[xr.Dataset, xr.DataArray],
    *,
    x: str,
    y: str,
    ax: "matplotlib.axes.Axes" = None,
    _collection: Optional[AgentCollection] = None,
    _domain_area: float = 1,
    marker: Union[Path, Patch, str, list, dict] = "auto",
    hue: str = None,
    orientation: Union[str, float, Sequence[float]] = None,
    size: Union[str, float, Sequence[float]] = None,
    size_norm: Union[str, dict, "matplotlib.colors.Normalize"] = None,
    size_vmin: float = None,
    size_vmax: float = None,
    size_scale: float = 0.0005,
    label: str = None,
    cmap: Union[str, dict] = None,
    norm: Union[str, dict] = None,
    vmin: Union[float, str] = None,
    vmax: Union[float, str] = None,
    cbar_label: str = None,
    cbar_labels: dict = None,
    add_colorbar: bool = None,
    cbar_kwargs: dict = None,
    **coll_kwargs,
) -> AgentCollection:
    """Draws agent positions onto the given axis.

    Data can be a :py:class:`~xarray.DataArray` or :py:class:`~xarray.Dataset`
    with ``x``, ``y`` and the other encodings ``hue``, ``orientation`` and
    ``size`` being strings that denote

    Args:
        d (Union[xarray.Dataset, xarray.DataArray]): The agent data.
            The recommended format is that of datasets, because it can hold
            more data variables.
            If using data *arrays*, multiple coordinates can be used to set the
            data. Also, the ``name`` attribute needs to be set (to avoid
            accidentally using the wrong data).
        x (str): Name of the data dimension or variable that holds x positions.
        y (str): Name of the data dimension or variable that holds y positions.
        ax (matplotlib.axes.Axes, optional): The axis to plot to; will use the
            current axis if not given.
        _collection (Optional[AgentCollection], optional): If a collection
            exists, will update this collection instead of creating a new one.
        _domain_area (float, optional): The domain area, used to normalize the
            marker sizes for a consistent visual appearance. This is only
            needed when initializing the collection and will set its
            ``size_scale`` parameter to ``size_scale * _domain_area``.
        marker (Union[matplotlib.path.Path, matplotlib.patches.Patch, str, list, dict], optional):
            Which marker to use for the agents. See
            :py:class:`.AgentCollection` for available arguments.
        hue (str, optional): The name of the data dimension or variable that
            holds the values from which marker colors are determined.
        orientation (Union[str, float, Sequence[float]], optional): The name of
            the data variable or dimension that holds the orientation data (in
            radians). If scalar or sequence, will pass those explicit values
            through.
        size (Union[str, float, Sequence[float]], optional): The name of the
            data variable or dimension that holds the values from which the
            marker size is determined.
        size_norm (Union[str, dict, matplotlib.colors.Normalize], optional):
            A normalization for the sizes.
            The :py:class:`~dantro.plot.utils.color_mngr.ColorManager` is used
            to set up this normalization, thus offering all the corresponding
            capabilities.
            Actual sizes are computed by `xarray.plot.utils._parse_size``.
        size_vmin (float, optional): The value that is mapped to the smallest
            size of the markers.
        size_vmax (float, optional): The value that is mapped to the largest
            size of the markers.
        size_scale (float, optional): A scaling factor that is applied to the
            size of all markers; use this to control overall size.
        label (str, optional): How to label the collection that is created.
            Will be used as default colorbar label and for the legend.
        cmap (Union[str, dict], optional): The colormap to use for the markers.
            Supports :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
        norm (Union[str, dict], optional): Norm to use for the marker colors.
            Supports :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
        vmin (Union[float, str], optional): The value of the ``hue`` dimension
            that is mapped to the beginning of the colormap.
        vmax (Union[float, str], optional): The value of the ``hue`` dimension
            that is mapped to the end of the colormap.
        cbar_labels (dict, optional): Colorbar labels, parsed by the
            :py:class:`~dantro.plot.utils.color_mngr.ColorManager`.
        add_colorbar (bool, optional): Whether to add a colorbar or not. Only
            used upon creating the collection.
        cbar_kwargs (dict, optional): Passed on to
            :py:meth:`~dantro.plot.utils.color_mngr.ColorManager.create_cbar`
        **coll_kwargs: Passed on to :py:class:`.AgentCollection`; see there
            for further available arguments.
    """
    import matplotlib.pyplot as plt

    def parse_sizes(d: np.ndarray, size_norm) -> np.ndarray:
        from xarray.plot.utils import _MARKERSIZE_RANGE, _parse_size

        # Use ColorManager to evaluate the size_norm argument
        size_norm = ColorManager(
            norm=size_norm, vmin=size_vmin, vmax=size_vmax
        ).norm

        # Compute and apply the size mapping.
        # Need a correction factor to let mapped sizes appear approximately
        # equally large as unmapped ones. With _parse_sizes returning values
        # within _MARKERSIZE_RANGE (typically [18, 72]), the correction is
        # simply to divide by a value of *approximately* that magnitude.
        size_mapping = _parse_size(d, size_norm)
        corr_factor = 1 / np.array(_MARKERSIZE_RANGE).mean()
        return corr_factor * size_mapping.loc[d.values.ravel()].values

    def get_orientations(d: Union[xr.DataArray, xr.Dataset]):
        """If ``orientation`` is a string, looks up the orientation data,
        otherwise simply passes the value through, letting the AgentManager
        handle the argument."""
        if isinstance(orientation, str):
            return _get_data(d, orientation)
        return orientation

    def get_xy(d: Union[xr.DataArray, xr.Dataset]) -> np.ndarray:
        """Retrieve and aggregate x and y position data"""
        xpos = _get_data(d, x)
        ypos = _get_data(d, y)
        return np.stack([xpos, ypos], -1)
        # equivalent to np.dstack([xpos, ypos])[0], but that one is slower

    # .........................................................................

    if isinstance(d, BaseDataContainer):
        d = d.data

    if not isinstance(d, (xr.Dataset, xr.DataArray)):
        raise TypeError(f"Expected xr.Dataset or xr.DataArray, got {type(d)}!")

    if ax is None:
        ax = plt.gca()

    # Depending on the collection existing or not, update it or create it.
    coll = _collection
    if coll is not None:
        # Only need to update the existing collection's offsets, orientations,
        # array data, sizes ...
        coll.set_offsets(get_xy(d))

        if orientation:
            coll.set_orientations(get_orientations(d))

        if hue and coll.has_markers:
            coll.set_array(_get_data(d, hue))
            if vmin is None or vmax is None:
                coll.autoscale()
                coll.set_clim(vmin, vmax)

        if size and isinstance(size, str):
            # TODO Make this more efficient?
            coll.set_markersizes(parse_sizes(_get_data(d, size), size_norm))

        return coll

    # else: Still need to create everything ...................................
    cm = ColorManager(
        cmap=cmap, norm=norm, vmin=vmin, vmax=vmax, labels=cbar_labels
    )

    # Prepare size data ...
    if size:
        if "sizes" in coll_kwargs:
            raise ValueError(
                "Cannot pass both `sizes` and `size` arguments! "
                "Use `size` to pass marker sizes via a data dimension, a "
                "sequence of sizes, or a scalar size (same for all agents)."
            )

        if isinstance(size, str):
            coll_kwargs["sizes"] = parse_sizes(_get_data(d, size), size_norm)

        else:
            # Let the collection take care of this
            coll_kwargs["sizes"] = size

    # Create collection
    # Need to pass orientations here already, because it determines whether
    # this will be a heterogeneous AgentCollection (with individual paths) or
    # a homogeneous one (with a single path, copied for each agent).
    coll = AgentCollection(
        get_xy(d),
        offset_transform=AffineDeltaTransform(ax.transData),
        orientations=get_orientations(d) if orientation else None,
        label=label,
        cmap=cm.cmap,
        norm=cm.norm,
        size_scale=(size_scale * _domain_area),
        marker=marker,
        **coll_kwargs,
    )
    if hue and coll.has_markers:
        coll.set_array(_get_data(d, hue))
        coll.autoscale()
        coll.set_clim(vmin, vmax)

    # Add collections to the current axis and ensure equal aspect ratio
    ax.set_aspect("equal")
    ax.add_collection(coll)
    if coll.tails is not None:
        ax.add_collection(coll.tails)

    # Create the colorbar
    cbar = None
    if add_colorbar or (add_colorbar is None and (hue and marker)):
        cbar = cm.create_cbar(
            coll,
            fig=ax.get_figure(),
            ax=ax,
            label=cbar_label,
            **ensure_dict(cbar_kwargs),
        )
        # NOTE matplotlib already adds it as `coll.colorbar` attribute

    return coll


# -----------------------------------------------------------------------------


@is_plot_func(use_dag=True, supports_animation=True)
def abmplot(
    *,
    data: Dict[str, Union[xr.Dataset, xr.DataArray]],
    hlpr: PlotHelper,
    to_plot: Dict[str, dict],
    frames: str = None,
    frames_isel: Union[int, slice] = None,
    domain: Union[str, dict, Tuple[float, float, float, float]] = "fixed",
    adjust_figsize_to_domain: bool = True,
    figsize_aspect_offset: float = 0.2,
    suptitle_fstr: str = "{dim:} = {val:5d}",
    suptitle_kwargs: dict = dict(fontfamily="monospace"),
    add_legend: bool = None,
    **shared_kwargs,
):
    """Plots agents in a domain, animated over time.

    Features:

    - Plot multiple "layers" of agents onto the same axis.
    - Encode ``hue``, ``size`` and ``orientation`` of all agents.
    - Keep track of position history through an animation and draw the history
      as (constant or decaying) tails.
    - Automatically determine the domain bounds, allowing to dynamically follow
      the agents through the domain.
    - Generate marker paths ad-hoc or use pre-defined markers that are relevant
      in the context of visualizing agent-based models, e.g. ``fish``.

    Example output:

    .. raw:: html

        <video width="720" src="../_static/_gen/abmplot/fish.mp4" controls></video>

    .. admonition:: Corresponding plot configuration
        :class: dropdown

        The following configuration was used to generate the above example
        animation.

        .. literalinclude:: ../../tests/cfg/plots/abm_plots.yml
            :language: yaml
            :start-after: ### START -- doc_fish
            :end-before: ### END ---- doc_fish

        The used dummy data (``circle_walk…``) is an
        :py:class:`xarray.Dataset` with data variables ``x``, ``y``,
        ``orientation``, each one spanning dimensions ``time`` and ``agents``.
        Data variables do not have coordinates in this case, but it would be
        possible to supply some.

    .. admonition:: Development roadmap

        The following extensions of this plot function are *planned*:

        - Subplot support, allowing to manually distribute `to_plot` entries
          onto individual subplots.
        - Faceting support, allowing to automatically distribute *all* data
          layers onto rows and columns and thus performing a dimensionality
          reduction on them.

    .. admonition:: See also

        - :py:func:`draw_agents`
        - :ref:`plot_funcs_abm`

    Args:
        data (Dict[str, Union[xarray.Dataset, xarray.DataArray]]): A dict
            holding the data that is to be plotted.
        hlpr (PlotHelper): The plot helper instance.
        to_plot (Dict[str, dict]): Individual plot specifications. Each entry
            refers to one "layer" of agents being plotted and its key needs to
            match the corresponding entry in ``data``.
            See :py:func:`.draw_agents` for available arguments.
            Note that the ``shared_kwargs`` also end up in the entries here.
        frames (str, optional): Which data dimension to use for the animation.
        frames_isel (Union[int, slice], optional): An index-selector that is
            applied on the ``frames`` dimension (*without* dropping empty
            dimensions resulting from this operation). Scalar values will be
            wrapped into a list of size one.
        domain (Union[str, dict, Tuple[float, float, float, float]], optional):
            Specification of the domain on which the agents are plotted.
            If a 4-tuple is given, it defines the position of the *fixed*
            ``(left, right, bottom, top)`` borders of the domain; a 2-tuple
            will be interpreted as ``(0, right, 0, top)``.
            If a string is given, it determines the *mode* by which the domain
            is determined. It can be ``fixed`` (determine bounds from all
            agent positions, then keep them constant) or ``follow`` (domain
            dynamically follows agents, keeping them centered).
            If a dict is given, it is passed to :py:func:`._set_domain`,
            and offers more options like ``pad`` or ``aspect``.
        adjust_figsize_to_domain (bool, optional): If True, will adjust the
            figure size to bring it closer to the aspect ratio of the domain,
            thus reducing whitespace.
        figsize_aspect_offset (float, optional): A constant offset that is
            added to the aspect ratio of the domain to arrive at the target
            aspect ratio of the figure used for ``adjust_figsize_to_domain``.
        suptitle_fstr (str, optional): The format string to use for the
            figure's super-title. If empty or None, no suptitle is added.
            The format string can include the following keys:
            ``dim`` (frames dimension name), ``val`` (current coordinate value,
            will be an index if no coordinates are available),
            ``coords`` (a dict containing all layers' iterator states),
            and ``frame_no`` (enumerates the frame number).
        suptitle_kwargs (dict, optional): Passed to
            :py:meth:`~matplotlib.figure.Figure.suptitle`.
        add_legend (bool, optional): Whether to add a legend.
            ⚠️ This is currently *experimental* and does not work as expected.
        **shared_kwargs: Shared keyword arguments for all plot specifications.
            Whatever is specified here is used as a basis for each individual
            layer in ``to_plot``, i.e. *updated* by the entries in ``to_plot``.
            See :py:func:`.draw_agents` for available arguments.
    """

    def prepare_layer_specs(
        name: str, *, shared_kwargs: dict, **plot_kwargs
    ) -> dict:
        """Parses arguments for each layer into a dict"""
        lyr = dict(
            name=name,
            data=None,
            plot_kwargs=None,
            ax=None,
            coll=None,
            frames=None,
            frames_iter=None,
            num_frames=None,
        )

        # Prepare plot kwargs
        plot_kwargs = recursive_update(
            copy.deepcopy(shared_kwargs), copy.deepcopy(plot_kwargs)
        )
        plot_kwargs["label"] = plot_kwargs.get("label", name)

        # Store it, exposing some parameters on the upper level
        lyr["plot_kwargs"] = plot_kwargs
        lyr["x"] = plot_kwargs["x"]
        lyr["y"] = plot_kwargs["y"]

        # Prepare data
        try:
            d = data[name]

        except KeyError as err:
            _avail = ", ".join(data.keys())
            raise ValueError(
                f"Missing data named '{name}'! Available:  {_avail}\n"
                "Check that the name is correctly defined. If you are using "
                "the DAG framework for data selection, make sure the tag is "
                "computed by including it into `compute_only` or setting the "
                "`force_compute` flag for the tagged node."
            ) from err

        # Extract parameters that are not meant to be passed on to the plot
        frames = plot_kwargs.pop("frames", None)
        frames_isel = plot_kwargs.pop("frames_isel", None)

        # Prepare the frames iterator
        if frames_isel is not None:
            if isinstance(frames_isel, int):
                selector = {frames: [frames_isel]}
            else:
                selector = {frames: frames_isel}
            try:
                d = d.isel(selector, drop=False)

            except Exception as exc:
                raise ValueError(
                    f"Failed applying index-selection for layer '{name}'! "
                    "Did you intend to apply a selection to this layer's "
                    "data? If not, set `frames_isel` to None.\n"
                    f"  frames:       {frames}\n"
                    f"  frames_isel:  {frames_isel}\n"
                    f"  → selector:   {selector}\n"
                    f"  data:\n{d}"
                ) from exc

        frames_iter = None if frames is None else d.groupby(frames)

        # ... and store it all to the layer specs
        lyr["frames"] = frames
        lyr["frames_iter"] = frames_iter
        lyr["num_frames"] = len(frames_iter) if frames_iter else 0
        lyr["data"] = d

        # (Mutably) evaluate vmin & vmax on hue and size dimensions
        if plot_kwargs.get("hue"):
            _parse_vmin_vmax(plot_kwargs, _get_data(d, plot_kwargs["hue"]))

        if plot_kwargs.get("size"):
            _parse_vmin_vmax(
                plot_kwargs,
                _get_data(d, plot_kwargs["size"]),
                vmin_key="size_vmin",
                vmax_key="size_vmax",
            )

        return lyr

    def set_suptitle(
        fstr: str, *, shared_dim: str, frame_coords: dict, **fstr_kwargs
    ) -> Tuple[str, float]:
        """Sets the suptitle, filling it with information about the currently
        plotted frame, e.g. its coordinate."""

        # Need to extract a shared coordinate value
        shared_coord = None
        coords = [v for v in frame_coords.values() if v is not None]
        if coords and all([c == coords[0] for c in coords]):
            shared_coord = coords[0]

        # TODO intelligently set fstr?

        if fstr and shared_dim:
            fstr_kwargs["dim"] = shared_dim
            fstr_kwargs["val"] = shared_coord
            fstr_kwargs["coords"] = frame_coords
            try:
                hlpr.fig.suptitle(
                    fstr.format(**fstr_kwargs),
                    **ensure_dict(suptitle_kwargs),
                )
            except Exception as exc:
                _fstr_kwargs = "\n".join(
                    f"    {k}:  {v}" for k, v in fstr_kwargs.items()
                )
                raise ValueError(
                    "Failed setting suptitle, probably due to a formatting "
                    "issue! Check the format string and the available "
                    "formatting values are compatible.\n"
                    f"  suptitle_fstr:  {repr(fstr)}\n"
                    f"  fstr_kwargs:\n{_fstr_kwargs}"
                ) from exc

        return shared_dim, shared_coord

    def parse_frame_coords(*iters, names: list) -> dict:
        """Extracts the current frame coordinate values from the state of the
        iterators. This is necessary because coordinate values may also be none
        if some layers are not iterated over.
        """

        def parse_coord(frit: Optional[tuple]):
            if frit is None:
                return None
            c, _ = frit
            return c

        coords = {
            names[i]: parse_coord(frit[0]) for i, frit in enumerate(iters)
        }
        # TODO check for equal values in case there *is* a coordinate?!

        return coords

    def parse_domain_kwargs(
        domain: Union[None, str, list, tuple, dict]
    ) -> dict:
        if domain is None:
            return dict()
        elif isinstance(domain, str):
            return dict(mode=domain)
        elif isinstance(domain, (list, tuple)):
            return dict(extent=domain, pad=0)
        elif isinstance(domain, dict):
            return domain

        raise TypeError(
            f"Got invalid type {type(domain)} for `domain` argument! "
            "Valid types are:  None, str, tuple, list, dict"
        )

    # .........................................................................

    # Work on a copy of the plot spec, to be on the safe side.
    to_plot = copy.deepcopy(to_plot)

    # Bring data into expected form and
    log.note("Preparing data for ABM plot ...")
    shared_kwargs = dict(
        frames=frames,
        frames_isel=frames_isel,
        **shared_kwargs,
    )
    layers = {
        k: prepare_layer_specs(k, shared_kwargs=shared_kwargs, **spec)
        for k, spec in to_plot.items()
    }
    layer_names = list(layers)

    # Check frames iteration
    num_frames = [layers[lyr]["num_frames"] for lyr in layer_names]
    if not all([n == 0 or n == num_frames[0] for n in num_frames]):
        _info = "\n".join(
            f"  {layer_names[i]:12}: {n:4d}" for i, n in enumerate(num_frames)
        )
        raise ValueError(
            "There was a mismatch in the length of the given frame iterators! "
            "Iterators need either be of the same length along the `frames` "
            "dimension or zero (if there is no `frames` dimension for that "
            f"layer).\nGiven iterator lengths were:\n{_info}"
        )
    else:
        num_frames = num_frames[0]

    # May need to switch in to / out of animation mode
    if num_frames > 1:
        hlpr.enable_animation()
    else:
        hlpr.disable_animation()

    # Construct the aggregated frames iterators object
    _frits = [layers[lyr]["frames_iter"] for lyr in layer_names]
    mock_frit = lambda: [None] * max(1, num_frames)
    frames_iters = zip(
        *[frit if frit is not None else mock_frit() for frit in _frits]
    )

    # Determine the name of a potentially existing shared frames dimension
    shared_frames_dim = None
    frame_dims = [layers[lyr]["frames"] for lyr in layer_names]
    unique_frame_dim_names = set([f for f in frame_dims if f])
    if unique_frame_dim_names and len(unique_frame_dim_names) == 1:
        shared_frames_dim = unique_frame_dim_names.pop()

    # Extract domain information and set the x- and y-axis limits.
    domain_area, _aspect = _set_domain(
        ax=hlpr.ax, layers=layers, **parse_domain_kwargs(domain)
    )

    # Inform about the plot ...
    log.remark("  Data variables:         %s", ", ".join(to_plot))
    log.remark(
        "  Frames dimension(s):    %s",
        shared_frames_dim
        if shared_frames_dim
        else ", ".join(f"{f}" for f in frame_dims),
    )
    log.remark(
        "  Number of frames:       %s",
        num_frames if num_frames > 1 else "(snapshot)",
    )
    log.remark(
        "  Domain area:            %.4g  (aspect: %.3g)", domain_area, _aspect
    )

    if adjust_figsize_to_domain:
        adjust_figsize_to_aspect(_aspect + figsize_aspect_offset, fig=hlpr.fig)
        _fw, _fh = hlpr.fig.get_size_inches()
        log.remark("  Figure size adjusted:   [%.3g, %.3g]", _fw, _fh)

    # .. Define single frame and animation update functions ...................
    # For simplicity, these use objects from the outer scope

    def plot_frame(frame_data: list, *, frame_no: int):
        """Plots a single frame from the given frame data"""
        frame_coords = parse_frame_coords(frame_data, names=layer_names)
        set_suptitle(
            suptitle_fstr,
            shared_dim=shared_frames_dim,
            frame_coords=frame_coords,
            frame_no=frame_no,
        )

        for lyr_no, coords_and_data in enumerate(frame_data):
            lyr_spec = layers[layer_names[lyr_no]]

            # Get the data
            if coords_and_data is None:
                # No iterator available; only need to draw once
                if lyr_spec["coll"] is not None:
                    continue
                xy = lyr_spec["data"]

            else:
                # always draw, using the data from the iterator
                _, xy = coords_and_data

            lyr_spec["coll"] = draw_agents(
                xy,
                ax=hlpr.ax,
                **lyr_spec["plot_kwargs"],
                _collection=lyr_spec["coll"],
                _domain_area=domain_area,
            )

        # May want to update the domain
        _set_domain(
            ax=hlpr.ax,
            layers=layers,
            update_only=True,
            **parse_domain_kwargs(domain),
        )

        hlpr.invoke_enabled()  # FIXME should not have to do this here!?

    def update():
        # First frame was already plotted, grab it directly
        yield

        # Plot the rest
        for frame_no, frame_data in enumerate(frames_iters):
            plot_frame(frame_data, frame_no=frame_no + 1)

            # Done with this frame; yield control to the animation framework
            # which will grab the frame...
            yield

        if frame_no > 1:
            log.info("Animation finished.")

    hlpr.register_animation_update(update)

    # .. Actual plotting ......................................................

    # Plot only the first frame here; all further frames are plotted via the
    # helper's animation framework.
    # What is done here is meant for the case of the snapshot and advances the
    # iterator by one. The captured iterator in the update function above
    # will then continue from the second frame.
    plot_frame(frames_iters.__next__(), frame_no=0)

    # Create the legend
    if add_legend:
        log.caution("Adding a legend via abmplot will probably be buggy!")
        hlpr.ax.legend()



"""Test the datagroup module"""

import networkx as nx
import pytest
from pkg_resources import resource_filename

import utopya
from utopya.eval.groups import GraphGroup, UniverseGroup
from utopya.testtools import ModelTest

from .. import get_cfg_fpath

GRAPH_MODEL = "MinimalModel"  # FIXME Not a graph model
GG_CFG_PATH = get_cfg_fpath("graphgroup_cfg.yml")

# -----------------------------------------------------------------------------


@pytest.mark.skip("No graph model available")
def test_graphgroup():
    """Test the dantro.GraphGroup integration into utopya. The full
    functionality is tested on dantro side."""
    mt = ModelTest(GRAPH_MODEL, test_file=__file__)
    mv, dm = mt.create_run_load(from_cfg=GG_CFG_PATH)

    for uni in dm["multiverse"].values():
        gg_static = uni["data/CopyMeGraph/g_static"]
        gg_dynamic = uni["data/CopyMeGraph/g_dynamic"]
        cfg = uni["cfg"]

        for gg in [gg_static, gg_dynamic]:

            # Check that it was loaded correctly. This requires that, on C++
            # side, the correct attributes were set and the model was built.
            assert isinstance(gg, GraphGroup)
            assert gg._GG_node_container == "_vertices"
            assert gg._GG_edge_container == "_edges"
            assert gg._GG_attr_directed == "is_directed"
            assert gg._GG_attr_parallel == "allows_parallel"
            assert (
                gg._GG_attr_edge_container_is_transposed
                == "edge_container_is_transposed"
            )
            assert gg._GG_WARN_UPON_BAD_ALIGN == True

            # Test that the graph can be created and properties can be added.
            g = gg.create_graph(at_time=0)

            if gg == gg_dynamic:
                g = gg.create_graph(
                    at_time_idx=-1,
                    node_props=["some_state", "some_trait"],
                    edge_props=["weights"],
                )
            else:
                g = gg.create_graph(
                    at_time_idx=-1, node_props=["some_state", "some_trait"]
                )

            # Check that the number of vertices matches
            assert (
                g.number_of_nodes()
                == cfg[GRAPH_MODEL]["create_graph"]["num_vertices"]
            )

            # Get the properties from the graph
            some_state = nx.get_node_attributes(g, "some_state")
            some_trait = nx.get_node_attributes(g, "some_trait")

            # Check that the node properties were correctly added
            for n in g.nodes():
                assert some_state[n] == gg["some_state"].isel(time=-1).sel(
                    vertex_idx=n
                )
                assert some_trait[n] == gg["some_trait"].isel(time=-1).sel(
                    vertex_idx=n
                )

            # Check that edge properties were added for the dynamic graph
            if gg == gg_dynamic:
                assert nx.get_edge_attributes(g, "weights")



"""Tests utopya_backend.base_model"""

import copy
import os
import random
import signal
import time

import numpy as np
import pytest

from utopya.testtools import ModelTest
from utopya.tools import load_yml, write_yml
from utopya_backend.model import BaseModel, StepwiseModel
from utopya_backend.signal import SIG_STOPCOND, SIGNAL_INFO

from .. import DUMMY_MODEL
from .._fixtures import *

# -- Fixtures -----------------------------------------------------------------


@pytest.fixture
def minimal_pspace_cfg(with_test_models) -> dict:
    """Returns the minimally necessary parameter space that will get passed to
    the model; does so by instantiating a test Multiverse and starting a run,
    then re-loading the file. This way, it is ensured to be in sync with
    whatever the Multiverse does.
    The model-specific information is stripped from it"""
    mtc = ModelTest(DUMMY_MODEL)
    mv = mtc.create_mv()
    mv.run()

    cfg = load_yml(os.path.join(mv.dirs["data"], "uni0", "config.yml"))
    assert os.path.exists(cfg["output_path"])

    # Drop model-specific content, to be re-added
    del cfg[cfg["root_model_name"]]
    del cfg["root_model_name"]

    # Need to remove the existing output file, otherwise the test model cannot
    # create its output file there.
    os.remove(cfg["output_path"])
    assert not os.path.exists(cfg["output_path"])
    assert os.path.isdir(os.path.dirname(cfg["output_path"]))

    # Yield to model
    # NOTE Important to yield instead of returning, because otherwise the
    #      Multiverse's temporary directory object goes out of scope and leads
    #      to removal of the above directiony, which would require that we
    #      emulate that ... which we don't want.
    yield cfg

    # Can do cleanup here, if necessary


# -- Implementations ----------------------------------------------------------


class MyModel(BaseModel):
    """A test model implementation from BaseModel. This doesn't do much..."""

    def setup(self, *, some_param, max_n_iter: int) -> None:
        self.some_param = some_param
        self.max_n_iter = max_n_iter

    def should_iterate(self) -> bool:
        """A method that determines whether :py:meth:`.iterate` should be
        called or not."""
        return self.n_iterations < self.max_n_iter

    def iterate(self) -> None:
        """Called repeatedly until the end of the simulation, which can be
        either due to"""
        pass

    def should_write(self) -> bool:
        """A method that determines whether :py:meth:`.write_data` should be
        called after an iteration or not."""
        return False

    def write_data(self) -> None:
        """Performs data writing if :py:meth:`.should_write` returned true."""
        pass


class MyStepwiseModel(StepwiseModel):
    """A model that tests StepwiseModel internals"""

    def setup(self, *, sleep_time: float, mock_signal: int = None):
        self._num_monitor_emits = 0
        self._num_writes = 0
        self._sleep_time = sleep_time
        self._total_sleep_time = 0.0
        self._mock_signal = mock_signal

        self.h5group.create_dataset("some_dset", (self.num_steps + 1,))

    def perform_step(self):
        dt = self._sleep_time + self.rng.uniform(0.01, 0.02)
        time.sleep(dt)
        self._total_sleep_time += dt

        if self._mock_signal is not None:
            self._signal_info = copy.deepcopy(SIGNAL_INFO)
            self._signal_info["got_signal"] = True
            self._signal_info["signum"] = self._mock_signal

    def write_data(self):
        self.h5group["some_dset"][self._num_writes] = self._num_writes
        self._num_writes += 1

    def trigger_monitor(self, **kwargs):
        super().trigger_monitor(**kwargs)
        self._num_monitor_emits += 1

    def monitor(self, monitor_info: dict) -> dict:
        monitor_info["total_sleep_time"] = self._total_sleep_time
        monitor_info["num_writes"] = self._num_writes
        return monitor_info


# -----------------------------------------------------------------------------


def test_BaseModel_init(minimal_pspace_cfg, tmpdir):
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyModel"
    cfg["MyModel"] = dict(some_param="foo", max_n_iter=13)

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Cannot instantiate if it's abstract
    with pytest.raises(TypeError, match="abstract class BaseModel"):
        BaseModel(cfg_file_path=cfg_path)

    # But this one works
    model = MyModel(cfg_file_path=cfg_path)

    # ... and has parameters carried over
    assert model.some_param == "foo"

    # Can also iterate
    assert model.n_iterations == 0
    model.run()
    assert model.n_iterations == 13

    # Teardown does not fail even if there is some unexpected error.
    # Here, provoke it by removing the attribute altogether
    model._h5file = None
    del model


def test_BaseModel_rngs(minimal_pspace_cfg, tmpdir):
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyModel"
    cfg["MyModel"] = dict(some_param="foo", max_n_iter=13)

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Set up a first model and let it generate random numbers
    N = 100
    m1 = MyModel(cfg_file_path=cfg_path)
    rands1 = dict(
        model=[m1.rng.random() for _ in range(N)],
        np=[np.random.random() for _ in range(N)],
        sys=[random.random() for _ in range(N)],
    )

    # When setting up a second model, it should generate the same numbers
    os.remove(cfg["output_path"])  # because m1's file still exists
    m2 = MyModel(cfg_file_path=cfg_path)
    rands2 = dict(
        model=[m2.rng.random() for _ in range(N)],
        np=[np.random.random() for _ in range(N)],
        sys=[random.random() for _ in range(N)],
    )

    assert np.isclose(rands1["model"], rands2["model"]).all()
    assert np.isclose(rands1["np"], rands2["np"]).all()
    assert np.isclose(rands1["sys"], rands2["sys"]).all()

    # When setting different seeds, this is not the case *for np & system*
    os.remove(cfg["output_path"])  # because m1s file still exists
    cfg["seed_numpy_rng"] = False
    cfg["seed_system_rng"] = None
    write_yml(cfg, path=cfg_path)

    m3 = MyModel(cfg_file_path=cfg_path)
    rands3 = dict(
        model=[m3.rng.random() for _ in range(N)],
        np=[np.random.random() for _ in range(N)],
        sys=[random.random() for _ in range(N)],
    )

    assert np.isclose(rands1["model"], rands3["model"]).all()
    assert not np.isclose(rands1["np"], rands3["np"]).any()
    assert not np.isclose(rands1["sys"], rands3["sys"]).any()


# -----------------------------------------------------------------------------


def test_StepwiseModel_basic(minimal_pspace_cfg, tmpdir):
    """Tests instantiation of a base model"""
    # Create the configuration file, filling it with model-specific info
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyStepwiseModel"
    cfg["MyStepwiseModel"] = dict(sleep_time=0.01)
    cfg["monitor_emit_interval"] = 0.1
    cfg["num_steps"] = 42

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Instantiate a model instance from that config file
    model = MyStepwiseModel(cfg_file_path=cfg_path)

    assert model.name == "MyStepwiseModel"
    assert model.time == 0
    assert model.write_start == 0
    assert model.write_every == 1
    assert model._num_writes == 1  # from initialization
    assert model._num_monitor_emits == 1  # from initialization

    # Run the model and test its state afterwards
    model.run()

    assert model.time == 42
    assert model._num_writes == 43
    assert model._num_monitor_emits >= 5

    # Hack around the time limit to run again and test interrupt handling
    model.should_write = lambda *a, **k: False
    model._time = 40
    model._mock_signal = signal.SIGINT
    with pytest.raises(SystemExit, match=str(128 + abs(signal.SIGINT))):
        model.run()

    model._time = 30
    model._mock_signal = signal.SIGTERM
    with pytest.raises(SystemExit, match=str(128 + abs(signal.SIGTERM))):
        model.run()

    model._time = 20
    model._mock_signal = SIG_STOPCOND
    with pytest.raises(SystemExit, match=str(128 + abs(SIG_STOPCOND))):
        model.run()

    model._time = 10
    model._mock_signal = 13
    with pytest.raises(SystemExit, match=str(128 + abs(13))):
        model.run()

    # Invoke some methods that are not accessible easily
    model._invoke_epilog(finished_run=False)
    model._invoke_epilog(finished_run=True)

    model.epilog = lambda *a, **k: 1 / 0
    with pytest.raises(ZeroDivisionError):
        model._invoke_epilog(finished_run=True)

    # h5py.File object should be closed after teardown
    h5file = model._h5file
    assert h5file
    del model
    assert not h5file


def test_StepwiseModel_missing_methods(minimal_pspace_cfg, tmpdir):
    """Tests incomplete model definition"""
    # Create the configuration file, filling it with model-specific info
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "my_model"
    cfg["my_model"] = dict()
    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    # Remains abstract and cannot be instantiated directly
    class IncompleteModel(StepwiseModel):
        pass

    with pytest.raises(TypeError, match="perform_step.*setup.*write_data"):
        IncompleteModel(cfg_file_path=cfg_path)

    # Can instantiate with those three methods implemented
    class NotIncompleteModel(StepwiseModel):
        def setup(self):
            pass

        def perform_step(self):
            pass

        def write_data(self):
            pass

    model = NotIncompleteModel(cfg_file_path=cfg_path)
    model.run()


def test_StepwiseModel_create_dset(minimal_pspace_cfg, tmpdir):
    """Tests creation of datasets"""
    cfg = minimal_pspace_cfg
    cfg["root_model_name"] = "MyStepwiseModel"
    cfg["MyStepwiseModel"] = dict(sleep_time=0.01)
    cfg["monitor_emit_interval"] = 0.1
    cfg["num_steps"] = 42
    cfg["write_start"] = 10
    cfg["write_every"] = 3

    # from which follows:
    num_writes = 11

    cfg_path = tmpdir.join("cfg.yml")
    write_yml(cfg, path=cfg_path)

    model = MyStepwiseModel(cfg_file_path=cfg_path)

    dset_one = model.create_ts_dset("one")
    assert dset_one.size == 0
    assert dset_one.shape == (0,)
    assert dset_one.maxshape == (num_writes,)

    dset_two = model.create_ts_dset("two", extra_dims=("x",), sizes=dict(x=10))
    assert dset_two.size == 0
    assert dset_two.shape == (0, 10)
    assert dset_two.maxshape == (num_writes, 10)

    dset_three = model.create_ts_dset(
        "three",
        extra_dims=(
            "x",
            "y",
        ),
        sizes=dict(x=2, y=3),
    )
    assert dset_three.size == 0
    assert dset_three.shape == (0, 2, 3)
    assert dset_three.maxshape == (num_writes, 2, 3)

    dset_four = model.create_ts_dset(
        "four", extra_dims=("x",), sizes=dict(x=10), coords=dict(x=range(10))
    )
    assert dset_four.size == 0
    assert dset_four.shape == (0, 10)
    assert dset_four.maxshape == (num_writes, 10)
    assert dset_four.attrs["coords_mode__x"] == "values"

    dset_five = model.create_ts_dset(
        "five",
        extra_dims=("x",),
        sizes=dict(x=10),
        coords=dict(x=dict(mode="trivial")),
    )
    assert dset_five.size == 0
    assert dset_five.shape == (0, 10)
    assert dset_five.maxshape == (num_writes, 10)
    assert dset_five.attrs["coords_mode__x"] == "trivial"

    # Errors
    with pytest.raises(ValueError, match="was not part of the list of"):
        model.create_ts_dset(
            "error1",
            extra_dims=("x",),
            sizes=dict(x=10),
            coords=dict(y="foo"),
        )

    with pytest.raises(ValueError, match="does not match"):
        model.create_ts_dset(
            "error2",
            extra_dims=("x",),
            sizes=dict(x=10),
            coords=dict(x=range(11)),
        )



"""Tests the utopya projects <...> subcommands of the CLI"""

import os

import pytest

import utopya
from utopya import PROJECTS

from .. import DEMO_DIR, TEST_PROJECT_NAME, get_cfg_fpath
from ..test_model_registry import tmp_cfg_dir, tmp_projects
from . import invoke_cli

VALID_INFO_FILE = get_cfg_fpath("project_info.yml")
INVALID_INFO_FILE = get_cfg_fpath("project_info_invalid.yml")

# -----------------------------------------------------------------------------


def test_list(tmp_projects):
    """Tests utopya projects ls"""
    # Lists projects as expected
    res = invoke_cli(("projects", "ls"))
    assert "Utopya Projects" in res.output
    assert TEST_PROJECT_NAME in res.output

    # Long mode shows additional info
    res = invoke_cli(("projects", "ls", "--long"))
    print(res.output)
    assert "metadata" in res.output
    assert "paths" in res.output

    assert res.output == invoke_cli(("projects", "ls", "-l")).output


def test_edit(tmp_projects, monkeypatch):
    """Tests utopya projects edit"""
    args = ("projects", "edit", TEST_PROJECT_NAME)
    res = invoke_cli(args, input="y\n")
    print(res.output)
    assert res.exit_code == 1
    assert "Editing project registry file failed!" in res.output  # bad editor

    monkeypatch.setenv("EDITOR", "echo")  # this will always work
    res = invoke_cli(args, input="y\n")
    assert res.exit_code == 0
    assert "Successfully edited project registry file" in res.output

    # Not continuing
    res = invoke_cli(args, input="N\n")
    assert res.exit_code == 0
    assert "Not opening" in res.output

    # Bad project name
    res = invoke_cli(("projects", "edit", "bad_project_name"), input="y\n")
    assert res.exit_code == 1
    assert "not found" in res.output


def test_remove(tmp_projects):
    """Tests utopya projects rm"""
    # Do not confirm
    res = invoke_cli(("projects", "rm", TEST_PROJECT_NAME), input="N\n")
    assert "Not removing" in res.output
    assert res.exit_code == 0

    # Skip confirmation but project name missing
    assert "missingProjectName" not in PROJECTS
    res = invoke_cli(("projects", "rm", "missingProjectName", "-y"))
    assert res.exit_code != 0
    assert "no such entry is present" in res.output
    assert "missingProjectName" not in PROJECTS

    # Now remove it
    assert TEST_PROJECT_NAME in PROJECTS
    res = invoke_cli(("projects", "rm", TEST_PROJECT_NAME, "-y"))
    assert res.exit_code == 0
    assert TEST_PROJECT_NAME not in PROJECTS


def test_register(tmp_projects):
    """Tests utopya project register"""
    assert TEST_PROJECT_NAME in PROJECTS

    reg_args = (
        "projects",
        "register",
        DEMO_DIR,
        "--custom-name",
        TEST_PROJECT_NAME,
    )

    # Invoke for the first time
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code == 0

    # Can invoke again, despite existing entry: Will validate by default
    res = invoke_cli(reg_args)
    print(res.output)
    assert res.exit_code == 0
    assert "Validating" in res.output
    assert f"Validation of entry '{TEST_PROJECT_NAME}' succeeded" in res.output

    # Explicitly pass path to (same) info file
    res = invoke_cli(
        reg_args
        + ("--info-file", os.path.join(DEMO_DIR, ".utopya-project.yml"))
    )
    print(res.output)
    assert res.exit_code == 0
    assert "Validating" in res.output

    # Explicitly pass exists_action
    res = invoke_cli(reg_args + ("--exists-action", "validate"))
    print(res.output)
    assert res.exit_code == 0

    # exists_action: raise
    res = invoke_cli(reg_args + ("--exists-action", "raise"))
    print(res.output)
    assert res.exit_code != 0
    assert "already exists" in res.output

    # exists_action: overwrite
    res = invoke_cli(reg_args + ("--exists-action", "overwrite"))
    print(res.output)
    assert res.exit_code == 0
    assert "Overwriting" in res.output

    # exists_action: update
    res = invoke_cli(reg_args + ("--exists-action", "update"))
    print(res.output)
    assert res.exit_code == 0
    assert "Updating" in res.output

    # Invalid exists_action
    res = invoke_cli(reg_args + ("--exists-action", "bad_exists_action"))
    print(res.output)
    assert res.exit_code == 2  # error comes from click
    assert "bad_exists_action" in res.output

    with pytest.raises(ValueError, match="Invalid `exists_action`"):
        PROJECTS.register(
            base_dir=DEMO_DIR,
            custom_project_name=TEST_PROJECT_NAME,
            exists_action="bad_exists_action",
        )

    # Store under custom name
    res = invoke_cli(reg_args + ("--custom-name", "some_custom_name"))
    print(res.output)
    assert res.exit_code == 0
    assert f"Successfully stored or updated 'some_custom_name'" in res.output

    # Name needs to match
    res = invoke_cli(
        reg_args
        + ("--custom-name", "some_custom_name", "--require-matching-names")
    )
    print(res.output)
    assert res.exit_code != 0
    assert f"does not match the name given in the project info" in res.output

    # Missing info file
    res = invoke_cli(("projects", "register", os.path.join(DEMO_DIR, "../")))
    print(res.output)
    assert res.exit_code != 0

    # --- Test also registering the models alongside the project
    res = invoke_cli(
        reg_args + ("--exists-action", "overwrite", "--with-models")
    )
    print(res.output)
    assert res.exit_code == 0
    assert "and 3 accompanying model(s)" in res.output

    # --- Use different files to test overwriting, updating, validating
    # Validation failure
    res = invoke_cli(
        reg_args
        + ("--info-file", VALID_INFO_FILE, "--custom-name", TEST_PROJECT_NAME)
    )
    print(res.output)
    assert res.exit_code != 0
    assert "Validating" in res.output
    assert "Their YAML diff is as follows" in res.output

    # ... old info is still there
    project = PROJECTS[TEST_PROJECT_NAME]
    assert "A demo project" in project.metadata.description

    # exists_action: overwrite ...
    res = invoke_cli(
        reg_args
        + (
            "--info-file",
            VALID_INFO_FILE,
            "--custom-name",
            TEST_PROJECT_NAME,
            "--exists-action",
            "overwrite",
        )
    )
    print(res.output)
    assert res.exit_code == 0
    assert "Overwriting" in res.output
    assert f"Successfully stored" in res.output

    # ... have new info there
    PROJECTS.reload()
    project = PROJECTS[TEST_PROJECT_NAME]
    assert "A TEST project" in project.metadata.description

    # exists_action: update ...
    # TODO

    # ... have updated info there now
    # TODO

    # Loading a file with an invalid data schema will fail
    res = invoke_cli(reg_args + ("--info-file", INVALID_INFO_FILE))
    print(res.output)
    assert res.exit_code != 0
    assert "1 validation error" in res.output
    assert "Extra inputs are not permitted" in res.output



"""This module holds functions used in the Multiverse's cluster mode"""

import re
from typing import List

# -----------------------------------------------------------------------------


def parse_node_list(node_list_str: str, *, mode: str, rcps: dict) -> List[str]:
    """Parses the node list to a list of node names and checks against the
    given resolved cluster parameters.

    Depending on ``mode``, different forms of the node list are parsable.
    For ``condensed`` mode:

    .. code-block::

        node042
        node[002,004-011,016]
        m05s[0204,0402,0504]
        m05s[0204,0402,0504],m08s[0504,0604,0701],m13s0603,m14s[0501-0502]
    """
    if not node_list_str:
        raise ValueError("Got empty node list string representation!")

    pattern = ""
    node_list = []

    if mode == "condensed":
        # See  https://regex101.com/r/yGdl2j/1  for deduction of regex pattern
        pattern = r"((?P<prefix>[\w\d]+)\[(?P<nodes>[\d\-\,\s]+)\]|(?P<node>[\w\d]+)),?"
        matches = re.finditer(pattern, node_list_str)

        for match in matches:
            if match["node"] is not None:
                # Was only a single node, not in condensed form
                node_list.append(match["node"])
                continue

            # Got a condensed match; need to continue parsing it ...
            prefix, nodes = match["prefix"], match["nodes"]

            # Remove whitespace and split
            segments = nodes.replace(" ", "").split(",")
            segments = [seg.split("-") for seg in segments]

            # Get the string width
            digits = len(segments[0][0])

            # In segments, lists longer than 1 are intervals, the
            # others are node numbers of a single node.
            # Expand intervals
            segments = [
                [int(seg[0])]
                if len(seg) == 1
                else list(range(int(seg[0]), int(seg[1]) + 1))
                for seg in segments
            ]

            # Combine to list of individual node numbers
            node_nos = []
            for seg in segments:
                node_nos += seg

            # Need the numbers as strings
            node_nos = [
                "{val:0{digs:}d}".format(val=no, digs=digits)
                for no in node_nos
            ]

            # Now, finally, parse the list
            node_list += [prefix + no for no in node_nos]

    else:
        raise ValueError(f"Invalid parser mode '{mode}'! Available: condensed")

    # Have node_list now. Make some consistency checks
    if rcps["num_nodes"] != len(node_list):
        raise ValueError(
            f"The parsed node list ({node_list}) has a different length "
            f"({len(node_list)}) than specified by the `num_nodes` parameter "
            f"({rcps['num_nodes']})! Make sure the string representation of "
            f"the node list ({repr(node_list_str)}) is parsable by the "
            f"selected parser mode '{mode}' "
            f"with the following regex pattern: {repr(pattern)}"
        )

    if rcps["node_name"] not in node_list:
        raise ValueError(
            f"The current node's `node_name` '{rcps['node_name']}' is not "
            f"part of the parsed node list: {', '.join(node_list)}"
        )

    # All ok. Sort and return
    return sorted(node_list)



"""Benchmarking tools for Models"""

import logging
import time
from collections import OrderedDict
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import h5py as h5
from dantro.tools import format_time as _format_time

log = logging.getLogger(__name__)

DEFAULT_TIMERS_ONE_SHOT: Tuple[str, ...] = (
    "init",
    "setup",
    "prolog",
    "run",
    "epilog",
    "teardown",
    "simulation",
)
"""Names of default one-shot timers in :py:class:`.ModelBenchmarkMixin`"""

DEFAULT_TIMERS_CUMULATIVE: Tuple[str, ...] = (
    "model_iteration",
    "monitor",
    "write_data",
    "full_iteration",
)
"""Names of default cumulative timers in :py:class:`.ModelBenchmarkMixin`"""


# -----------------------------------------------------------------------------


class Timer:
    """Implements a simple timer that can be paused and continued."""

    _name: str
    _one_shot: bool
    _time_func: Callable = time.time

    _running: bool
    _latest: float
    _elapsed: float
    _finished: bool

    def __init__(
        self,
        name: str,
        *,
        time_func: Callable = None,
        one_shot: bool = False,
        start: bool = False,
    ):
        self._name = name
        self._one_shot = one_shot
        self.reset()

        if time_func is not None:
            self._time_func = time_func

        if start:
            self.start()

    def _get_time(self) -> float:
        return self._time_func()

    def _assert_not_finished(self):
        if self.finished:
            raise RuntimeError(
                f"Tried to update timer '{self}' that was already marked "
                "as finished."
            )

    def __str__(self) -> str:
        segments = []
        segments.append(f"Timer '{self.name}'")
        segments.append(f"{self.elapsed:.3g}s elapsed")
        if self.running:
            segments.append("running")
        if self.finished:
            segments.append("finished")
        return f"<{', '.join(segments)}>"

    def start(self):
        self._assert_not_finished()
        self._unpause()

    def pause(self) -> float:
        self._assert_not_finished()

        if not self.running:
            raise RuntimeError(f"Cannot pause already paused timer {self}!")

        self._elapsed += self._get_time() - self._latest
        self._latest = None
        self._running = False

        if self.one_shot:
            self._finished = True
        return self.elapsed

    def unpause(self):
        if self.one_shot:
            raise RuntimeError(
                f"{self} is a one-shot timer and cannot be unpaused!"
            )
        self._assert_not_finished()
        self._unpause()

    def _unpause(self):
        self._latest = self._get_time()
        self._running = True

    def stop(self) -> float:
        self._assert_not_finished()
        if self.running:
            self.pause()
        self._finished = True
        return self.elapsed

    def reset(self):
        self._latest = None
        self._running = False
        self._finished = False
        self._elapsed = 0.0

    @property
    def name(self) -> str:
        return self._name

    @property
    def running(self) -> bool:
        return self._running

    @property
    def finished(self) -> bool:
        return self._finished

    @property
    def elapsed(self) -> float:
        if self._running:
            return self._elapsed + (self._get_time() - self._latest)
        return self._elapsed

    @property
    def one_shot(self) -> bool:
        return self._one_shot


# -----------------------------------------------------------------------------


class ModelBenchmarkMixin:
    """A mixin class that allows to conveniently gather information on the run
    time that individual parts of the model iteration require and also store it
    in the model's dataset.

    To use this, simply inherit it into your model class definition:

    .. testcode::

        from utopya_backend import BaseModel, ModelBenchmarkMixin

        class MyModel(ModelBenchmarkMixin, BaseModel):
            pass

    By default, this will enable the benchmarking and will both show the
    result at the end of the run as well as write it to a separate benchmarking
    group in the default HDF5 data group.
    To further configure its behaviour, add a ``benchmark`` entry to your
    model's configuration. For available parameters and default values, refer
    to :py:meth:`._configure_benchmark`.
    """

    _timers: Dict[str, Timer]

    _TIMER_FALLBACK_RV: Any = -1
    """The fallback value that is returned by :py:meth:`.pause_timer` and
    :py:meth:`.stop_timer` when benchmarking is completely disabled.
    """

    _dgrp_bench: Optional[h5.Group] = None
    _dset_total: Optional[h5.Dataset] = None
    _dset_cumulative: Optional[h5.Dataset] = None
    __dgrp_name: str
    __dset_dtype: str
    __dset_compression: int
    _dset_cumulative_invocation_times: List[int]

    __enabled: bool = True
    __write: bool = None
    _show_on_exit: bool = True
    _add_time_elapsed_to_monitor_info: bool = False
    _time_elapsed_info_fstr: str

    # TODO consider not having default values here

    # .........................................................................

    def __init__(self, *args, **kwargs):
        # Start with default values and timers
        self._timers = OrderedDict()
        self._add_default_timers()

        self.start_timer("simulation")
        self.start_timer("init")
        super().__init__(*args, **kwargs)
        self.stop_timer("init")

        # Have the configuration available only now, after init (running setup)
        self._configure_benchmark(**self._bench_cfg)

        # Find out if the class this is mixed-in to has a step-based iteration
        # scheme, in which case some procedures may run differently.
        self._is_stepwise_model = hasattr(self, "write_start") and hasattr(
            self, "write_every"
        )

        if self.__enabled:
            # Create the group that benchmark data will be written to
            if self.__write:
                self._dgrp_bench = self.h5group.create_group(self.__dgrp_name)

            self.log.info("Model benchmarking set up.")
        else:
            self.log.debug("Model benchmarking disabled.")

    def _add_default_timers(self):
        self.add_one_shot_timers(*DEFAULT_TIMERS_ONE_SHOT)
        self.add_cumulative_timers(*DEFAULT_TIMERS_CUMULATIVE)

    def _configure_benchmark(
        self,
        *,
        enabled: bool = True,
        show_on_exit: bool = True,
        add_to_monitor: bool = False,
        write: bool = True,
        group_name: str = "benchmark",
        compression: int = 3,
        dtype: str = "float32",
        info_fstr: str = "  {name:>15s} : {time_str:s}",
    ):
        """Applies benchmark configuration parameters.

        Args:
            enabled (bool, optional): Whether to enable benchmarking. If False,
                the behaviour will be exactly the same, but timer invocations
                will simply be ignored.

                .. note::

                    Despite being disabled, a very minor performance hit can
                    still be expected (a few booleans that are evaluated).
                    Only removing the mixin altogether will alleviate that.

            show_on_exit (bool, optional): Whether to print an info-level log
                message at the end of the simulation, showing elapsed times.
            add_to_monitor (bool, optional): Whether to add elapsed times to
                the monitoring data.
            write (bool, optional): Whether to write data to HDF5 dataset.
                The cumulative timers are stored at each invocation of
                ``write_data``, while the one-shot timers are only written at
                the end of a simulation run.
            group_name (str, optional): The name of the HDF5 group to nest the
                output datasets in.
            compression (int, optional): HDF5 compression level.
            dtype (str, optional): HDF5 data type for timing information.
                By default, this is reduced float precision, because the times
                given by :py:func:`time.time` are not as precise anyway.
            info_fstr (str, optional): The format string to use for generation
                of :py:meth:`.elapsed_info`.
                Available keys: ``name``, ``seconds`` (float), ``time_str``
                (pre-formatted using :py:func:`dantro.tools.format_time`).
        """
        self.__enabled = enabled
        self._add_time_elapsed_to_monitor_info = add_to_monitor
        self._show_time_elapsed_on_exit = show_on_exit
        self._time_elapsed_info_fstr = info_fstr

        self.__write = write
        self.__dset_compression = compression
        self.__dset_dtype = dtype
        self.__dgrp_name = group_name

        if not self.__enabled:
            # Some timers have already started; easiest way is to just reset
            # all of them so that they behave effectively as disabled.
            for t in self.timers.values():
                t.reset()

    # .. Adding timers ........................................................

    def add_one_shot_timers(self, *names, **kwargs):
        for name in names:
            self._add_timer(name, one_shot=True, **kwargs)

    def add_cumulative_timers(self, *names, **kwargs):
        for name in names:
            self._add_timer(name, one_shot=False, **kwargs)

    def _add_timer(self, name, *, one_shot: bool, **kwargs):
        self.timers[name] = Timer(name, one_shot=one_shot, **kwargs)
        return self.timers[name]

    # .. Controlling timers ...................................................

    @property
    def timers(self) -> Dict[str, Timer]:
        return self._timers

    def _get_timer(self, name: str) -> Timer:
        try:
            return self.timers[name]
        except KeyError as err:
            _avail = ", ".join(sorted(self.timers))
            raise ValueError(
                f"No benchmark timer named '{name}' was added!\n"
                f"Available timers:  {_avail}"
            ) from err

    def start_timer(self, name: str) -> None:
        if not self.__enabled:
            return

        self._get_timer(name).start()

    def pause_timer(self, name: str) -> Union[float, Any]:
        if not self.__enabled:
            return self._TIMER_FALLBACK_RV

        return self._get_timer(name).pause()

    def unpause_timer(self, name: str) -> None:
        if not self.__enabled:
            return

        self._get_timer(name).unpause()

    def stop_timer(self, name: str) -> Union[float, Any]:
        if not self.__enabled:
            return self._TIMER_FALLBACK_RV

        return self._get_timer(name).stop()

    # .. Retrieving timer data ................................................

    @property
    def elapsed(self) -> Dict[str, float]:
        return {k: t.elapsed for k, t in self._timers.items()}

    @property
    def elapsed_cumulative(self) -> Dict[str, float]:
        return {
            k: t.elapsed for k, t in self._timers.items() if not t.one_shot
        }

    @property
    def elapsed_one_shot(self) -> Dict[str, float]:
        return {k: t.elapsed for k, t in self._timers.items() if t.one_shot}

    @property
    def elapsed_info(self) -> str:
        """Prepares a formatted string with all elapsed times"""
        return "\n".join(
            self._time_elapsed_info_fstr.format(
                name=name,
                seconds=seconds,
                time_str=_format_time(seconds, ms_precision=2),
            )
            for name, seconds in self.elapsed.items()
        )

    # .. Storing timer data ...................................................

    def _write_dset_total(self):
        if not self.__enabled or not self.__write:
            return

        elapsed = self.elapsed
        N = len(elapsed)

        # May still need to create the dataset
        if self._dset_total is None:
            ds = self._dgrp_bench.create_dataset(
                "total",
                (N,),
                maxshape=(N,),
                chunks=True,
                compression=self.__dset_compression,
                dtype=self.__dset_dtype,
            )

            ds.attrs["dim_names"] = ["label"]
            ds.attrs["coords_mode__label"] = "values"
            ds.attrs["coords__label"] = list(elapsed.keys())

            self._dset_total = ds

        # TODO check what happens if invoked repeatedly, possibly with new
        #      timers added in between

        self._dset_total[:] = list(elapsed.values())

    def _write_dset_cumulative(self):
        if not self.__enabled or not self.__write:
            return

        elapsed_cumulative = self.elapsed_cumulative
        N = len(elapsed_cumulative)

        # May still need to create it
        if self._dset_cumulative is None:
            ds = self._dgrp_bench.create_dataset(
                "cumulative",
                (0, N),
                maxshape=(None, N),
                chunks=True,
                compression=self.__dset_compression,
                dtype=self.__dset_dtype,
            )
            ds.attrs["dim_names"] = ["n_iterations", "label"]

            if not self._is_stepwise_model:
                # As constantly updating write times attribute would be too
                # costly, denote the times as trivial indices for now and
                # later update that attribute (at the very end) using the list
                # containing invocation times (in number of iterations) that
                # is built up meanwhile.
                ds.attrs["coords_mode__n_iterations"] = "trivial"
            else:
                ds.attrs["coords_mode__n_iterations"] = "start_and_step"
                _sas = [self.write_start, self.write_every]
                ds.attrs["coords__n_iterations"] = _sas

            ds.attrs["coords_mode__label"] = "values"
            ds.attrs["coords__label"] = list(elapsed_cumulative.keys())

            self._dset_cumulative = ds
            self._dset_cumulative_invocation_times = []

        # May need to expand size along time dimension
        ds = self._dset_cumulative
        ds.resize(ds.shape[0] + 1, axis=0)

        # Now write:
        ds[-1, :] = list(elapsed_cumulative.values())

        # Extend list of write times (to be written to attribute at the end
        # of the run)
        if not self._is_stepwise_model:
            self._dset_cumulative_invocation_times.append(self.n_iterations)

    # .. Inject into simulation procedure .....................................
    # Note that __init__ also contains a timer

    def _invoke_setup(self):
        self.start_timer("setup")
        self._bench_cfg = self.cfg.pop("benchmark", {})
        super()._invoke_setup()
        self.stop_timer("setup")

    def _pre_run(self):
        self.start_timer("run")
        super()._pre_run()

    def _post_run(self, *, finished_run: bool):
        super()._post_run(finished_run=finished_run)

        # Stop all remaining timers
        self.stop_timer("run")
        for timer in self.timers.values():
            if timer.finished or timer.name == "teardown":
                continue
            self.stop_timer(timer.name)

        # Write total values for all timers
        self._write_dset_total()

        # Ensure that coordinate labels for n_iterations are stored
        if (
            self.__enabled
            and self.__write
            and self._dset_cumulative is not None
            and self._dset_cumulative_invocation_times
            and not self._is_stepwise_model
        ):
            ds = self._dset_cumulative
            times = self._dset_cumulative_invocation_times
            ds.attrs["coords_mode__n_iterations"] = "values"
            ds.attrs["coords__n_iterations"] = times
            self.log.debug("times:\n%s", times)
            self.log.debug("ds.attrs:  %s", dict(ds.attrs.items()))

        # Show times
        if self.__enabled and self._show_on_exit:
            self.log.info(
                "Elapsed times for parts of this simulation:\n\n%s\n",
                self.elapsed_info,
            )

    def _invoke_prolog(self):
        self.start_timer("prolog")
        super()._invoke_prolog()
        self.stop_timer("prolog")

    def _pre_iterate(self):
        self.unpause_timer("full_iteration")
        super()._pre_iterate()

    def _invoke_iterate(self):
        self.unpause_timer("model_iteration")
        super()._invoke_iterate()
        self.pause_timer("model_iteration")

    def _pre_monitor(self):
        self.unpause_timer("monitor")
        super()._pre_monitor()

    def _emit_monitor(self):
        if self._add_time_elapsed_to_monitor_info:
            self._monitor_info["timers"] = self.elapsed
        super()._emit_monitor()

    def _post_monitor(self):
        super()._post_monitor()
        self.pause_timer("monitor")

    def _invoke_write_data(self):
        self.unpause_timer("write_data")
        super()._invoke_write_data()
        self._write_dset_cumulative()
        self.pause_timer("write_data")

    def _post_iterate(self):
        super()._post_iterate()
        self.pause_timer("full_iteration")

    def _invoke_epilog(self, **kwargs):
        self.start_timer("epilog")
        super()._invoke_epilog(**kwargs)
        self.stop_timer("epilog")

    def __del__(self):
        self.start_timer("teardown")
        super().__del__()
        self.stop_timer("teardown")



"""Module that coordinates utopya's persistent config directory"""

import logging
import os

from ._yaml import load_yml, write_yml

log = logging.getLogger(__name__)

# Some globally relevant variables --------------------------------------------
#
# NOTE For CLI support, some of these are duplicated in utopya_cli._shared
#      When doing updates here, also update them there.

UTOPYA_CFG_DIR = os.path.expanduser("~/.config/utopya")
"""Path to the persistent utopya configuration directory"""

UTOPYA_CFG_FILE_NAMES = dict(
    user="user_cfg.yml",
    utopya="utopya_cfg.yml",
    batch="batch_cfg.yml",
)
"""Names and paths of valid configuration entries"""

UTOPYA_CFG_FILE_PATHS = {
    k: os.path.join(UTOPYA_CFG_DIR, fname)
    for k, fname in UTOPYA_CFG_FILE_NAMES.items()
}
"""Absolute configuration file paths"""

UTOPYA_CFG_SUBDIR_NAMES = dict(
    models="models",
    projects="projects",
)
"""Names and paths of valid configuration subdirectories"""

UTOPYA_CFG_SUBDIRS = {
    k: os.path.join(UTOPYA_CFG_DIR, dirname)
    for k, dirname in UTOPYA_CFG_SUBDIR_NAMES.items()
}
"""Absolute configuration file paths"""

PROJECT_INFO_FILE_SEARCH_PATHS = (
    ".utopya_project.yml",
    ".utopya-project.yml",
)
"""Potential names of project info files, relative to base directory"""


# -----------------------------------------------------------------------------


def get_cfg_path(cfg_name: str) -> str:
    """Returns the absolute path to the specified configuration file"""
    try:
        return UTOPYA_CFG_FILE_PATHS[cfg_name]

    except KeyError as err:
        _avail = ", ".join(UTOPYA_CFG_FILE_NAMES.keys())
        raise ValueError(
            f"No configuration entry '{cfg_name}' available! "
            f"Possible keys: {_avail}"
        ) from err


def load_from_cfg_dir(cfg_name: str) -> dict:
    """Load a configuration file; returns empty dict if no file exists.

    Args:
        cfg_name (str): The name of the configuration to read

    Returns:
        dict: The configuration as read from the config directory; if no file
            is available, will return an empty dict.
    """
    cfg_fpath = get_cfg_path(cfg_name)
    try:
        d = load_yml(cfg_fpath)

    except FileNotFoundError:
        log.debug(
            "No '%s' configuration file exists at %s ! Returning empty.",
            cfg_name,
            cfg_fpath,
        )
        return dict()

    # If the yaml object is None for whatever reason, return an empty dict
    if d is None:
        return dict()
    return d


def write_to_cfg_dir(cfg_name: str, obj: dict):
    """Writes a YAML represetation of the given object to the configuration
    directory. Always overwrites a possibly existing file.

    Args:
        cfg_name (str): The configuration name
        obj (dict): The yaml-representable object that is to be written;
            usually a dict.
    """
    write_yml(obj, path=get_cfg_path(cfg_name))



"""This module implements the ModelRegistryEntry class, which contains a set
of ModelInfoBundle objects. These can be combined in a labelled and unlabelled
fashion.
"""

import copy
import logging
import os
import time
from itertools import chain
from typing import Generator, Iterator, Tuple, Union

from .._yaml import load_yml, write_yml
from ..cfg import UTOPYA_CFG_DIR
from ..exceptions import (
    BundleExistsError,
    BundleValidationError,
    MissingBundleError,
    ModelRegistryError,
)
from ..tools import pformat
from .info_bundle import TIME_FSTR, ModelInfoBundle

log = logging.getLogger(__name__)


# -----------------------------------------------------------------------------


class ModelRegistryEntry:
    """Holds model config bundles for a single model.

    Instances of this class are associated with a file in the model registry
    directory. Upon change, they update that file.
    """

    def __init__(self, model_name: str, *, registry_dir: str):
        """Initialize a model registry entry from a registry directory.

        Expects a file named <model_name>.yml in the registry directory. If it
        does not exist, creates an empty one.
        """
        self._model_name = model_name
        self._registry_dir = registry_dir

        self._bundles = dict()
        self._default_label = None

        # If no file exists for this entry, create empty one. Otherwise: load.
        if not os.path.exists(self.registry_file_path):
            log.debug(
                "Creating empty registry file for registry entry of "
                "model '%s' ...",
                self.model_name,
            )
            self._update_registry_file()  # creates file if path doesn't exist

        else:
            self._load_from_registry()

        log.debug(
            "Initialized registry entry for model '%s' with a total "
            "of %d bundles.",
            self.model_name,
            len(self),
        )

    @property
    def model_name(self) -> str:
        """Name of this model"""
        return self._model_name

    @property
    def registry_dir(self) -> str:
        """The associated registry directory"""
        return self._registry_dir

    @property
    def registry_file_path(self) -> str:
        """The absolute path to the registry file"""
        return os.path.join(self.registry_dir, f"{self.model_name}.yml")

    @property
    def default_label(self) -> Union[str, None]:
        """The default label"""
        return self._default_label

    @default_label.setter
    def default_label(self, val: Union[str, None]):
        """Sets the default label value. If None, there will be no default."""
        self.set_default_label(val, update_registry_file=True)

    @property
    def default_bundle(self) -> ModelInfoBundle:
        """Returns the default bundle, if set; raises if not"""
        if self.default_label is not None:
            return self._bundles[self.default_label]

        raise ModelRegistryError(
            f"No `default_label` set for {self}; "
            "cannot return a default bundle!"
        )

    # Magic methods ...........................................................

    def __len__(self) -> int:
        """Returns number of registered bundles"""
        return len(self._bundles)

    def __contains__(self, other: Union[ModelInfoBundle, str]) -> bool:
        """Checks if the given object *or* key is part of this registry entry."""
        if isinstance(other, str):
            return other in self._bundles
        return other in self.values()

    def __str__(self) -> str:
        return "<{} '{}'; {} bundle{}, default: '{}'>".format(
            type(self).__name__,
            self.model_name,
            len(self),
            "s" if len(self) != 1 else "",
            self.default_label,
        )

    def __eq__(self, other) -> bool:
        """Check for equality by inspecting stored bundles and model name"""
        if not isinstance(other, ModelRegistryEntry):
            return False
        return self._as_dict(self) == other._as_dict(other)

    # Access ..................................................................

    def __getitem__(self, key: str) -> ModelInfoBundle:
        """Return a bundle for the given label. If None, tries to return the
        single registered item.
        """
        try:
            return self._bundles[key]

        except KeyError as err:
            _avail = ", ".join(self.keys())
            raise MissingBundleError(
                f"No bundle labelled '{key}' registered in {self}!\n"
                f"Available labels:  {_avail}"
            ) from err

    def item(self) -> ModelInfoBundle:
        """Retrieve a single bundle for this model, if not ambiguous.

        This will work only in two cases:
            - If a default label is set, returns the corresponding bundle
            - If there is only a single bundle availble, returns that one

        Returns:
            ModelInfoBundle: The unambiguously selectable model info bundle

        Raises:
            ModelRegistryError: In case the selection was ambiguous
        """
        log.remark("Getting info bundle for model '%s' ...", self.model_name)
        if self.default_label is not None:
            log.remark("  ... using default label:  %s", self.default_label)
            return self.default_bundle

        elif len(self) != 1:
            _avail = ", ".join(self.keys())
            raise MissingBundleError(
                f"Could not unambiguously select single bundle from {self}, "
                "because no default was set or because the number of bundles "
                "is not exactly one.\n"
                "Define a `default_label` or use `__getitem__` to access a "
                f"bundle with a specific label (available: {_avail})."
            )

        label = next(iter(self.keys()))
        log.remark("  ... selecting only registered info bundle:  %s", label)
        return self[label]

    def keys(self) -> Iterator[str]:
        """Returns keys for item access, i.e.: all registered keys"""
        return self._bundles.keys()

    def values(self) -> Generator[ModelInfoBundle, None, None]:
        """Returns stored model config bundles, starting with labelled ones"""
        for k in self.keys():
            yield self[k]

    def items(
        self,
    ) -> Generator[Tuple[Union[str, int], ModelInfoBundle], None, None]:
        """Returns keys and registered bundles, starting with labelled ones"""
        for k in self.keys():
            yield k, self[k]

    # Manipulation ............................................................

    def add_bundle(
        self,
        *,
        label: str,
        set_as_default: bool = None,
        exists_action: str = "raise",
        update_registry_file: bool = True,
        **bundle_kwargs,
    ) -> ModelInfoBundle:
        """Add a new configuration bundle to this registry entry.

        This makes sure that the added bundle does not compare equal to an
        already existing bundle.

        Args:
            label (str): The label under which to add it.
            set_as_default (bool, optional): Controls whether to set this
                bundle's ``label`` as the default label for this entry.
            exists_action (str, optional): What to do if the given ``label``
                already exists:

                * ``raise``: Do not add a new bundle and raise (default)
                * ``skip``: Do not add a new bundle, instead return the exist
                * ``overwrite``: Overwrite the existing bundle
                * ``validate``: Make sure that the new bundle compares equal
                    to the one that already exists.

            update_registry_file (bool, optional): Whether to write changes
                directly to the registry file.
            **bundle_kwargs: Passed on to construct the ``ModelInfoBundle``
                that is to be stored.

        Raises:
            BundleExistsError: If ``label`` already exists and
                ``exists_action`` was set to ``raise``.
            BundleValidationError: If ``validate`` was given but the bundle
                already stored under the given ``label`` does not compare
                equal to the to-be-added bundle.
        """
        EXISTS_ACTIONS = ("raise", "skip", "overwrite", "validate")

        if not isinstance(label, str):
            raise TypeError(
                "The label for a model info bundle needs to be a string, "
                f"but was {type(label)} {label}!"
            )

        if exists_action not in EXISTS_ACTIONS:
            raise ValueError(
                f"Invalid `exists_action` '{exists_action}'! "
                f"Possible values: {', '.join(EXISTS_ACTIONS)}"
            )

        bundle = ModelInfoBundle(model_name=self.model_name, **bundle_kwargs)

        if label in self:
            if exists_action == "skip":
                log.caution(
                    "A bundle labelled '%s' already exists; not adding.", label
                )
                return self[label]

            elif exists_action == "validate":
                if self[label] != bundle:
                    # Generate a diff such that its clearer where they differ
                    import difflib

                    diff = "\n".join(
                        difflib.Differ().compare(
                            pformat(self[label].as_dict).split("\n"),
                            pformat(bundle.as_dict).split("\n"),
                        )
                    )

                    raise BundleValidationError(
                        f"Bundle validation failed for label '{label}'! "
                        "The to-be-added bundle did not compare equal to the "
                        "bundle that already exists under that label.\n"
                        "Either change the `exists_action` argument to "
                        "'overwrite' or make sure the bundles are equal; "
                        f"their diff is as follows:\n\n{diff}"
                    )

                log.debug("Validation successful for label '%s'.", label)
                return self[label]

            elif exists_action == "overwrite":
                pass

            else:  # "raise"
                raise BundleExistsError(
                    f"An info bundle with label '{label}' already exists in "
                    f"{self}!\n"
                    "Set `exists_action` to 'overwrite', 'skip', or "
                    "'validate' to handle this error."
                )

        log.debug(
            "Adding bundle '%s' for model '%s' ...", label, self.model_name
        )
        self._bundles[label] = bundle

        if set_as_default:
            self.set_default_label(label, update_registry_file=False)

        if update_registry_file:
            self._update_registry_file()

        return bundle

    def pop(
        self, key: str, *, update_registry_file: bool = True
    ) -> ModelInfoBundle:
        """Pop a configuration bundle from this entry."""
        log.debug(
            "Removing bundle '%s' from registry entry for model '%s' ...",
            key,
            self.model_name,
        )
        bundle = self._bundles.pop(key)

        if key == self.default_label:
            self.set_default_label(None, update_registry_file=False)

        if update_registry_file:
            self._update_registry_file()

        return bundle

    def clear(self, *, update_registry_file: bool = True):
        """Removes all configuration bundles from this entry."""
        self._bundles = dict()
        self.default_label = None

        if update_registry_file:
            self._update_registry_file()

    def set_default_label(
        self, label: str, *, update_registry_file: bool = True
    ):
        """Sets the default label

        Args:
            label (str): The new label
            update_registry_file (bool, optional): Whether to update the
                registry file.
        """
        if label is not None and label not in self.keys():
            _avail = ", ".join(self.keys())
            raise ValueError(
                f"Given info bundle label '{label}' for the "
                f"'{self.model_name}' model does not exist and thus cannot be "
                "set as default!\n"
                f"Available labels:  {_avail}"
            )

        self._default_label = label
        log.debug("Set default label to '%s'", self.default_label)

        if update_registry_file:
            self._update_registry_file()

    # Loading and Storing .....................................................

    def _load_from_registry(self):
        """Load the YAML registry entry for this model from the associated
        registry file path.
        """
        try:
            obj = load_yml(self.registry_file_path)

        except Exception as exc:
            raise type(exc)(
                "Failed loading model registry file "
                f"from {self.registry_file_path}!"
            ) from exc

        # Loaded successfully
        # If a model name is given, make sure it matches the file name
        if self.model_name != obj.get("model_name", self.model_name):
            raise ValueError(
                f"Mismatch between expected model name '{self.model_name}' "
                f"and the model name '{obj['model_name']}' specified in the "
                f"registry file at {self.registry_file_path}! "
                "Check the file name and the registry file and make sure the "
                "model name matches exactly."
            )

        # Populate self. Need not update because content is freshly loaded.
        for label, kwargs in obj.get("info_bundles", {}).items():
            self.add_bundle(label=label, **kwargs, update_registry_file=False)

        self.set_default_label(
            obj.get("default_label"), update_registry_file=False
        )

    def _update_registry_file(self, *, overwrite_existing: bool = True) -> str:
        """Stores a YAML representation of this bundle in a file in the given
        directory and returns the full path.

        The file is saved under ``<model_name>.yml``, preserving the case of
        the model name.
        Before saving, this makes sure that no file exists in that directory
        whose lower-case version would compare equal to the lower-case version
        of this model.
        """
        if not os.path.exists(self.registry_dir):
            os.makedirs(self.registry_dir)

        fpath = self.registry_file_path
        fname = os.path.basename(fpath)

        # Check for duplicates, including lower case
        lc_duplicates = [
            fn
            for fn in os.listdir(self.registry_dir)
            if fn.lower() == fname.lower()
        ]
        if lc_duplicates and not overwrite_existing:
            _duplicates = ", ".join(lc_duplicates)
            raise FileExistsError(
                "At least one file with a file name conflicting with "
                f"'{fname}' was found in {self.registry_dir}: {_duplicates}. "
                "Manually delete or rename the conflicting "
                "file(s), taking care to not mix cases."
            )

        # Write to separate location; move only after write was successful.
        write_yml(self, path=f"{fpath}.tmp")
        os.replace(f"{fpath}.tmp", fpath)
        # NOTE fpath need not exist for os.replace to work

        log.debug("Successfully stored %s at %s.", self, fpath)

    # YAML representation .....................................................

    @classmethod
    def _as_dict(cls, obj) -> dict:
        """Return a copy of the dict representation of this object"""
        d = dict(
            model_name=obj.model_name,
            info_bundles=obj._bundles,
            default_label=obj.default_label,
        )
        return copy.deepcopy(d)

    @classmethod
    def to_yaml(cls, representer, node):
        """Creates a YAML representation of the data stored in this entry.
        As the data is a combination of a dict and a sequence, instances of
        this class are also represented as such.

        Args:
            representer (ruamel.yaml.representer): The representer module
            node (ModelRegistryEntry): The node to represent, i.e. an instance
                of this class.

        Returns:
            A YAML representation of the given instance of this class.
        """
        d = cls._as_dict(node)

        # Add some more information to it
        d["time_of_last_change"] = time.strftime(TIME_FSTR)

        # Return a YAML representation of the data
        return representer.represent_data(d)



"""The :py:class:`.WorkerManager` is a central part of utopya in that it
spawns and controls the tasks (:py:class:`~utopya.task.WorkerTask`) that are
to be worked on.
"""

import copy
import logging
import os
import queue
import sys
import time
import warnings
from datetime import datetime as dt
from typing import Callable, Dict, List, Sequence, Set, Union

from .exceptions import *
from .reporter import WorkerManagerReporter
from .stop_conditions import SIG_STOPCOND, StopCondition
from .task import SIGMAP, TaskList, WorkerTask
from .tools import format_time

log = logging.getLogger(__name__)

STOPCOND_EXIT_CODES: Sequence[int] = (
    -SIGMAP[SIG_STOPCOND],
    SIGMAP[SIG_STOPCOND],
    128 + abs(SIGMAP[SIG_STOPCOND]),
)
"""Exit codes of a :py:class:`~utopya.task.WorkerTask` that will be interpreted
as stemming from a stop condition. This depends on the signal used for stop
conditions (:py:data:`utopya.stop_conditions.SIG_STOPCOND`).
This sequence of possible exit codes takes into account that the sign may be
switched (depending on whether a signed or unsigned integer convention is
used) or where a convention is used such that a *handled* signal is turned into
an exit code of ``128 + abs(signum)``."""

# -----------------------------------------------------------------------------


class WorkerManager:
    """The WorkerManager class orchestrates :py:class:`~utopya.task.WorkerTask`
    objects: setting them up, invoking them, tracking their progress, and
    starting new workers if previous workers finished.

    Attrs:
        rf_spec (dict): The report format specifications that are used
            throughout the WorkerManager. These are invoked at different points
            of the operation of the WorkerManager: ``while_working``,
            ``after_work``, ``after_abort``, ``task_spawn``, ``task_finished``.
    """

    pending_exceptions: queue.Queue = None
    """A (FiFo) queue of :py:exc:`Exception` objects that will be handled by
    the :py:class:`.WorkerManager` during working. This is the interface that
    allows for other threads that have access to the WorkerManager to add an
    exception and let it be handled in the main thread.
    """
    rf_spec: dict = None
    """The report format specifications that are used throughout the
    WorkerManager. These are invoked at different points of the operation:

    - ``while_working``
    - ``after_work``
    - ``after_abort``
    - ``task_spawn``
    - ``task_finished``
    """

    times: dict = None
    """Holds profiling information"""

    def __init__(
        self,
        num_workers: Union[int, str] = "auto",
        poll_delay: float = 0.05,
        spawn_rate: int = -1,
        lines_per_poll: int = 50,
        periodic_task_callback: int = None,
        QueueCls: type = queue.Queue,
        reporter: WorkerManagerReporter = None,
        rf_spec: Dict[str, Union[str, List[str]]] = None,
        save_streams_on: Sequence[str] = (),
        nonzero_exit_handling: str = "ignore",
        interrupt_params: dict = None,
        cluster_mode: bool = False,
        resolved_cluster_params: dict = None,
    ):
        """Initialize the worker manager.

        Args:
            num_workers (Union[int, str], optional): The number of workers
                that can work in parallel. If 'auto' (default), uses
                os.cpu_count(). If below zero, deduces abs(num_workers) from
                the CPU count.
            poll_delay (float, optional): How long (in seconds) the delay
                between worker polls should be. For too small delays (<0.01),
                the CPU load will become significant.
            spawn_rate (int, optional): How many workers to spawn each working
                loop iteration. If -1, will assign new tasks to all free
                workers.
            lines_per_poll (int, optional): How many lines to read from
                each stream during polling of the tasks. This value should not
                be too large, otherwise the polling is delayed by too much.
                By setting it to -1, all available lines are read.
            periodic_task_callback (int, optional): If given, an additional
                task callback will be invoked after every
                ``periodic_task_callback`` poll events.
            QueueCls (type, optional): Which class to use for the queue.
                Defaults to the FiFo :py:class:`queue.Queue`.
            reporter (WorkerManagerReporter, optional): The reporter associated
                with this WorkerManager, reporting on the progress.
            rf_spec (Dict[str, Union[str, List[str]]], optional): The names of
                report formats that should be invoked at different points of
                the WorkerManager's operation.
                Possible keys: ``before_working``, ``while_working``,
                ``after_work``, ``after_abort``, ``task_spawn``,
                ``task_finished``. All other keys are ignored.

                The values of the dict can be either strings or lists of
                strings, where the strings always refer to report formats
                registered with the WorkerManagerReporter. This argument
                updates the default report format specifications.
            save_streams_on (Sequence[str], optional): On which events to
                invoke :py:meth:`~utopya.task.WorkerTask.save_streams`
                *during* work.
                Should be a sequence containing one or both of the keys
                ``on_monitor_update``, ``periodic_callback``.
            nonzero_exit_handling (str, optional): How to react if a WorkerTask
                exits with a non-zero exit code. For 'ignore', nothing happens.
                For 'warn', a warning is printed and the last 5 lines of the
                log are shown. For 'raise', the last 20 lines of the log is
                shown, all other tasks are terminated, and the WorkerManager
                exits with the same exit code as the WorkerTask exited with.
                Note that 'warn' will not lead to any messages if the worker
                died by SIGTERM, which presumable originated from a fulfilled
                stop condition. Use 'warn_all' to also receive warnings in
                this case.
            interrupt_params (dict, optional): Parameters that determine how
                the WorkerManager behaves when receiving KeyboardInterrupts
                during working.
                Possible keys:

                    ``send_signal``: Which signal to send to the workers. Can
                        be SIGINT (default), SIGTERM, SIGKILL, or any valid
                        signal as integer.
                    ``grace_period``: how long to wait for the other workers to
                        gracefully shut down. After this period (in seconds),
                        the workers will be killed via SIGKILL. Default is 5s.
                    ``exit``: whether to sys.exit at the end of start_working.
                        Default is True.

            cluster_mode (bool, optional): Whether similar tasks to those that
                are managed by this WorkerManager are, at the same time, worked
                on by other WorkerManager. This is relevant because the output
                of files might be affected by whether another WorkerManager
                instance is currently working on the same output directory.
                Also, in the future, this argument might be used to communicate
                between nodes.
            resolved_cluster_params (dict, optional): The corresponding cluster
                parameters.

        Raises:
            ValueError: For too negative ``num_workers`` argument
        """
        log.progress("Initializing WorkerManager ...")

        # Initialize attributes, some of which are property-managed
        self._num_workers = None
        self._poll_delay = None
        self._spawn_rate = None
        self._periodic_callback = periodic_task_callback
        self._tasks = TaskList()
        self._task_q = QueueCls()
        self._active_tasks = []
        self.stopped_tasks = TaskList()
        self._stop_conditions = set()
        self._reporter = None
        self._num_finished_tasks = 0
        self._nonzero_exit_handling = None
        self.save_streams_on = save_streams_on
        self._suppress_rf_specs = []
        self.pending_exceptions = queue.Queue()

        # Hand over arguments
        self.poll_delay = poll_delay
        self.spawn_rate = spawn_rate
        self.lines_per_poll = lines_per_poll
        self.nonzero_exit_handling = nonzero_exit_handling
        self._cluster_mode = cluster_mode
        self._resolved_cluster_params = resolved_cluster_params
        self.interrupt_params = interrupt_params if interrupt_params else {}
        self.num_workers = num_workers

        # Reporter-related, setting default rf_spec
        if reporter:
            self.reporter = reporter

        self.rf_spec = dict(
            before_working="while_working",
            while_working="while_working",
            task_spawned="while_working",
            task_finished="while_working",
            after_work="after_work",
            after_abort="after_work",
        )
        if rf_spec:
            self.rf_spec.update(rf_spec)

        # Provide some information
        log.note("  Number of available CPUs:  %d", os.cpu_count())
        log.note("  Number of workers:         %d", self.num_workers)
        log.note("  Non-zero exit handling:    %s", self.nonzero_exit_handling)

        # Store some profiling information
        self.times = dict(
            init=dt.now(), start_working=None, timeout=None, end_working=None
        )
        # These are also accessed by the reporter

        log.progress("Initialized WorkerManager.")

    # Properties ..............................................................
    @property
    def tasks(self) -> TaskList:
        """The list of all tasks."""
        return self._tasks

    @property
    def task_queue(self) -> queue.Queue:
        """The task queue."""
        return self._task_q

    @property
    def task_count(self) -> int:
        """Returns the number of tasks that this manager *ever* took care of.
        Careful: This is NOT the current number of tasks in the queue!
        """
        return len(self.tasks)

    @property
    def num_workers(self) -> int:
        """The number of workers that may work in parallel"""
        return self._num_workers

    @num_workers.setter
    def num_workers(self, val: Union[int, str]):
        """Set the number of workers that can work in parallel.

        Can use the string ``auto`` to use the CPU count for setting the number
        of workers. Additionally, negative integers will be summed with the
        number of available CPU cores.

        .. note::

            Will not allow to set values that are smaller than 1; any argument
            that would lead to such a value will be silently clipped to 1.
        """
        if val == "auto":
            self._num_workers = os.cpu_count()

        elif not isinstance(val, int):
            raise TypeError(
                "Expected integer or string 'auto' for setting `num_workers`, "
                f"but got {type(val).__name__} {repr(val)}!"
            )

        elif val < 0:
            self._num_workers = max(1, os.cpu_count() + val)

        else:
            self._num_workers = max(1, val)

        log.debug("Set number of workers to %d", self.num_workers)
        if self.num_workers > os.cpu_count():
            warnings.warn(
                f"Set WorkerManager to use more workers ({self.num_workers}) "
                f"than there are CPU cores ({os.cpu_count()}) on this machine",
                UserWarning,
            )

    @property
    def active_tasks(self) -> List[WorkerTask]:
        """The list of currently active tasks.

        Note that this information might not be up-to-date; a process might
        quit just after the list has been updated.
        """
        return self._active_tasks

    @property
    def num_finished_tasks(self) -> int:
        """The number of finished tasks. Incremented whenever a task leaves
        the active_tasks list, regardless of its exit status.
        """
        return self._num_finished_tasks

    @property
    def num_free_workers(self) -> int:
        """Returns the number of free workers."""
        return self.num_workers - len(self.active_tasks)

    @property
    def poll_delay(self) -> float:
        """Returns the delay between two polls"""
        return self._poll_delay

    @poll_delay.setter
    def poll_delay(self, val) -> None:
        """Set the poll delay to any positive value. For low values, a warning
        will be emitted.
        """
        if val <= 0.0:
            raise ValueError(f"Poll delay needs to be positive, was {val}!")
        elif val < 0.01:
            warnings.warn(
                "Setting a poll delay of {} < 0.01s can lead to "
                "significant CPU load. Consider choosing a higher "
                "value.",
                UserWarning,
            )
        self._poll_delay = val

    @property
    def spawn_rate(self) -> int:
        """Returns the spawn rate"""
        return self._spawn_rate

    @spawn_rate.setter
    def spawn_rate(self, val: int) -> None:
        """Set the spawn rate."""
        if not (val == -1 or val > 0) or not isinstance(val, int):
            raise ValueError(
                f"Spawn rate needs to be a positive integer or -1, was {val}!"
            )
        self._spawn_rate = val

    @property
    def stop_conditions(self) -> Set[StopCondition]:
        """All stop conditions that were ever passed to
        :py:meth:`~utopya.workermanager.WorkerManager.start_working` during the
        life time of this WorkerManager.
        """
        return self._stop_conditions

    @property
    def nonzero_exit_handling(self) -> str:
        """Behavior upon a worker exiting with a non-zero exit code.

        - with ``ignore``, nothing happens
        - with ``warn``, a warning is printed
        - with ``raise``, the log is shown and the WorkerManager exits
          with the same exit code as the corresponding
          :py:class:`~utopya.task.WorkerTask` exited with.
        """
        return self._nonzero_exit_handling

    @nonzero_exit_handling.setter
    def nonzero_exit_handling(self, val: str):
        """Set the ``nonzero_exit_handling`` attribute.

        Args:
            val (str): The value to set it to. Can be: ``ignore``, ``warn``,
                ``raise``.

        Raises:
            ValueError: For invalid value
        """
        allowed_vals = ("ignore", "warn", "warn_all", "raise")
        if val not in allowed_vals:
            raise ValueError(
                f"`nonzero_exit_handling` needs to be one of {allowed_vals}, "
                f"but was '{val}'."
            )

        self._nonzero_exit_handling = val

    @property
    def reporter(self) -> Union[WorkerManagerReporter, None]:
        """The associated :py:class:`~utopya.reporter.WorkerManagerReporter`
        or None, if no reporter is set.
        """
        return self._reporter

    @reporter.setter
    def reporter(self, reporter):
        """Set the Reporter object for this WorkerManager.

        This includes a check for the correct type and whether the reporter was
        already set.
        """
        if not isinstance(reporter, WorkerManagerReporter):
            raise TypeError(
                "Need a WorkerManagerReporter for reporting from "
                f"WorkerManager, got {type(reporter)}."
            )
        elif self.reporter:
            raise RuntimeError("Already set the reporter; cannot change it.")

        self._reporter = reporter

        log.debug("Set reporter of WorkerManager.")

    @property
    def cluster_mode(self) -> bool:
        """Returns whether the WorkerManager is in cluster mode"""
        return self._cluster_mode

    @property
    def resolved_cluster_params(self) -> dict:
        """Returns a copy of the cluster configuration with all parameters
        resolved (thus making some additional keys available on the top level).
        It is returned as a deep copy to avoid mutability issues.
        """
        return copy.deepcopy(self._resolved_cluster_params)

    # Public API ..............................................................

    def add_task(
        self, *, TaskCls: type = WorkerTask, **task_kwargs
    ) -> WorkerTask:
        """Adds a task to the WorkerManager.

        Args:
            TaskCls (type, optional): The WorkerTask-like type to use
            **task_kwargs: All arguments needed for WorkerTask initialization.
                See :py:class:`utopya.task.WorkerTask` for all valid arguments.

        Returns:
            WorkerTask: The created WorkerTask object
        """
        # Evaluate save_stream callbacks
        save_streams = lambda t: t.save_streams(final=False)

        if "monitor_updated" in self.save_streams_on:
            save_streams_on_monitor_update = True

        if "periodic_callback" in self.save_streams_on:
            periodic_callback = save_streams
        else:
            # Do nothing
            periodic_callback = lambda _: None

        # Define the function needed to calculate the task's progress
        def calc_progress(task) -> float:
            """Uses the task's stream objects to calculate the progress.

            If no stream objects are available, returns 0.
            """
            # Check if parsed objects were available
            if not task.outstream_objs:
                return 0.0

            # Extract the `progress` key from the latest entry; if it is not
            # available, use the same behaviour as above and return 0
            return task.outstream_objs[-1].get("progress", 0.0)

        # Prepare the callback functions needed by the reporter . . . . . . . .
        def task_spawned(task):
            """Performs action after a task was spawned.

            - checks if stream-forwarding was activated
            - invokes the 'task_spawned' report specification
            """
            # As the task might have been configured to forward streams, it
            # needs to be checked whether this would clash with the reporter's
            # output to stdout.
            # First, check if that was already taken care of
            if self.reporter and not self.reporter.suppress_cr:
                # Nope -> need to check if the task will forward streams
                if task.streams.get("out") and task.streams["out"]["forward"]:
                    # Yes -> need to suppress carriage returns by reporter and
                    # reduce report invokations by the WorkerManager main loop
                    self.reporter.suppress_cr = True
                    self._suppress_rf_specs.append("while_working")

            # Invoke the report
            self._invoke_report("task_spawned", force=True)

        def task_finished(task):
            """Performs actions after a task has finished.

            - invokes the ``task_finished`` report specification
            - registers the task with the reporter, which extracts information
              on the run time of the task and its exit status
            - in debug mode, performs an action upon non-zero task exit status
            """
            if self.reporter is not None:
                self.reporter.register_task(task)

            self._invoke_report("task_finished", force=True)

            # If there was a (non-zero) exit and the handling mode is set
            # accordingly, generate an exception and add it to the list of
            # pending exceptions. Handle exit codes that result from a stop
            # condition being fulfilled separately.
            if self.nonzero_exit_handling != "ignore" and task.worker_status:
                if task.worker_status in STOPCOND_EXIT_CODES:
                    exc = WorkerTaskStopConditionFulfilled(task)
                else:
                    exc = WorkerTaskNonZeroExit(task)
                self.pending_exceptions.put_nowait(exc)

        def monitor_updated(task):
            """Performs actions when there was a parsed object in the task's
            stream, i.e. when the monitor got an update.
            """
            if save_streams_on_monitor_update:
                save_streams(task)
            self._invoke_report("monitor_updated")

        # Prepare the arguments for the WorkerTask . . . . . . . . . . . . . .

        callbacks = dict(
            # Invoked by task itself
            spawn=task_spawned,
            finished=task_finished,
            parsed_object_in_stream=monitor_updated,
            #
            # Invoked by WorkerManager
            periodic=periodic_callback,
        )

        # Generate the WorkerTask-like object from the given parameters
        task = TaskCls(
            callbacks=callbacks, progress_func=calc_progress, **task_kwargs
        )

        # Append it to the task list and put it into the task queue
        self.tasks.append(task)
        self.task_queue.put_nowait(task)

        log.debug("Task %s (uid: %s) added.", task.name, task.uid)
        return task

    def start_working(
        self,
        *,
        detach: bool = False,
        timeout: float = None,
        stop_conditions: Sequence[StopCondition] = None,
        post_poll_func: Callable = None,
    ) -> None:
        """Upon call, all enqueued tasks will be worked on sequentially.

        Args:
            detach (bool, optional): If False (default), the WorkerManager
                will block here, as it continuously polls the workers and
                distributes tasks.
            timeout (float, optional): If given, the number of seconds this
                work session is allowed to take. Workers will be aborted if
                the number is exceeded. Note that this is not measured in CPU
                time, but the host systems wall time.
            stop_conditions (Sequence[StopCondition], optional): During the
                run these StopCondition objects will be checked
            post_poll_func (Callable, optional): If given, this is called after
                all workers have been polled. It can be used to perform custom
                actions during a the polling loop.

        Raises:
            NotImplementedError: if ``detach`` was set
            ValueError: For invalid (i.e., negative) timeout value
            WorkerManagerTotalTimeout: Upon a total timeout
        """
        if detach:
            raise NotImplementedError(
                "It is currently not possible to detach the WorkerManager "
                "from the main thread."
            )

        self._invoke_report("before_working")

        log.progress("Preparing to work ...")

        # Determine timeout arguments
        if timeout:
            if timeout <= 0:
                raise ValueError(
                    f"Invalid value for argument `timeout`: {timeout} -- "
                    "needs to be positive."
                )

            # Already calculate the time after which a timeout would be reached
            self.times["timeout"] = time.time() + timeout

        # Set the variable needed for checking; if above condition was not
        # fulfilled, this will be None
        timeout_time = self.times["timeout"]

        # Set some variables needed during the run
        poll_no = 0
        self.times["start_working"] = dt.now()

        # Prepare stop conditions, creating objects if needed
        if stop_conditions:
            stop_conditions = [
                sc if isinstance(sc, StopCondition) else StopCondition(**sc)
                for sc in stop_conditions
            ]

        # Inform about timeout and stop conditions
        if timeout:
            _to_fstr = "%X" if timeout < 60 * 60 * 12 else "%X, %d.%m.%y"
            _to_at = dt.fromtimestamp(timeout_time).strftime(_to_fstr)
        log.note(
            "  Timeout:         %s",
            "None" if not timeout else f"{_to_at} (in {format_time(timeout)})",
        )
        log.note(
            "  Stop conditions: %s",
            "None"
            if not stop_conditions
            else ", ".join([sc.name for sc in stop_conditions]),
        )

        log.hilight("Starting to work ...")

        # Keep track of the stop condition objects
        if stop_conditions:
            self._stop_conditions.update(set(stop_conditions))

        # Start with the polling loop
        # Basically all working time will be spent in there ...
        try:
            while self.active_tasks or self.task_queue.qsize() > 0:
                # Check total timeout
                if timeout_time is not None and time.time() > timeout_time:
                    raise WorkerManagerTotalTimeout()

                # Check if there was another reason for exiting
                self._handle_pending_exceptions()

                # Assign tasks to free workers
                self._assign_tasks(self.num_free_workers)

                # Do stream-related actions for each task
                for task in self.active_tasks:
                    # Read the streams and forward them (if the tasks were
                    # configured to do so)
                    task.read_streams(
                        forward_directly=True,
                        max_num_reads=self.lines_per_poll,
                    )
                    task.forward_streams()

                # Invoke a report
                self._invoke_report("while_working")

                # Check stop conditions
                if stop_conditions:
                    # Compile a list of workers where the stop conditions
                    # were fulfilled and store them.
                    fulfilled = self._check_stop_conds(stop_conditions)
                    self.stopped_tasks += fulfilled

                    # Now signal those workers such that they terminate.
                    self._signal_workers(fulfilled, signal=SIG_STOPCOND)

                # Poll the workers. (Will also remove no longer active workers)
                self._poll_workers()

                # Call the post-poll function
                if post_poll_func is not None:
                    log.debug(
                        "Calling post_poll_func %s ...",
                        post_poll_func.__name__,
                    )
                    post_poll_func()

                # Invoke periodic callback for all tasks
                if (
                    self._periodic_callback
                    and poll_no % self._periodic_callback == 0
                ):
                    self._invoke_periodic_callbacks()

                # Some information
                poll_no += 1
                log.debug(
                    "Poll # %6d:  %d active tasks",
                    poll_no,
                    len(self.active_tasks),
                )

                # Delay the next poll
                time.sleep(self.poll_delay)

            # Finished working
            # Handle any remaining pending exceptions
            self._handle_pending_exceptions()

        except (KeyboardInterrupt, WorkerManagerTotalTimeout) as exc:
            # Got interrupted or a total timeout.
            # Interrupt the workers, giving them some time to shut down ...

            # Suppress reporter to use CR; then inform via log messages
            if self.reporter:
                self.reporter.suppress_cr = True

            log.warning("Received %s.", type(exc).__name__)

            # Extract parameters from config
            # Which signal to send to workers
            signal = self.interrupt_params.get("send_signal", "SIGINT")

            # The grace period within which the tasks have to shut down
            grace_period = self.interrupt_params.get("grace_period", 5.0)

            # Send the signal
            log.info(
                "Sending signal %s to %d active task(s) ...",
                signal,
                len(self.active_tasks),
            )
            self._signal_workers(self.active_tasks, signal=signal)

            # Continuously poll them for a certain grace period in order to
            # find out if they have shut down.
            log.warning(
                "Allowing %s for %d task(s) to shut down ... "
                "(Ctrl + C to kill them now.)",
                format_time(grace_period, ms_precision=1),
                len(self.active_tasks),
            )
            grace_period_start = time.time()

            try:
                while True:
                    self._poll_workers()

                    # Check whether to leave the loop
                    if not self.active_tasks:
                        break
                    elif time.time() > grace_period_start + grace_period:
                        raise KeyboardInterrupt

                    # Need to continue. Delay polling ...
                    time.sleep(self.poll_delay)

            except KeyboardInterrupt:
                log.critical(
                    "Killing workers of %d tasks now ...",
                    len(self.active_tasks),
                )
                self._signal_workers(self.active_tasks, signal="SIGKILL")

                # Wait briefly (killing shouldn't take long), then poll
                # one last time to update all task's status.
                time.sleep(0.5)
                self._poll_workers()

            # Store end time and invoke a report
            self.times["end_working"] = dt.now()
            self._invoke_report("after_abort", force=True)

            log.hilight("Work session ended.")

            if type(exc) is KeyboardInterrupt and self.interrupt_params.get(
                "exit", True
            ):
                # Exit with appropriate exit code (128 + abs(signum))
                log.warning("Exiting after KeyboardInterrupt ...")
                sys.exit(128 + abs(SIGMAP[signal]))

            # Otherwise, just return control to the calling scope

        except WorkerManagerError as err:
            # Some error not related to the non-zero exit code occurred.
            # Gracefully terminate remaining tasks before re-raising the
            # exception

            # Suppress reporter to use CR; then inform via log messages
            if self.reporter:
                self.reporter.suppress_cr = True

            log.critical("Did not finish working! See above for error log.")

            # Now terminate the remaining active tasks
            log.hilight("Terminating active tasks ...")
            self._signal_workers(self.active_tasks, signal="SIGTERM")

            # Store end time and invoke a report
            self.times["end_working"] = dt.now()
            self._invoke_report("after_abort", force=True)

            # For some specific error types, do not raise but exit with status
            if isinstance(err, WorkerTaskNonZeroExit):
                log.critical("Exiting now ...")
                sys.exit(err.task.worker_status)

            # Some other error occurred; just raise
            log.critical("Re-raising error ...")
            raise

        # Register end time and invoke final report
        self.times["end_working"] = dt.now()
        self._invoke_report("after_work", force=True)

        print("")
        log.progress(
            "Finished working. Total tasks worked on: %d", self.task_count
        )

    # Non-public API ..........................................................

    def _invoke_report(self, rf_spec_name: str, *args, **kwargs):
        """Helper function to invoke the reporter's report function"""
        if self.reporter is None:
            return

        # Check whether to suppress this rf_spec
        if self._suppress_rf_specs and rf_spec_name in self._suppress_rf_specs:
            # Do not report this one
            return

        # Resolve the spec name
        rfs = self.rf_spec[rf_spec_name]

        if not isinstance(rfs, list):
            rfs = [rfs]

        for rf in rfs:
            self.reporter.report(rf, *args, **kwargs)

    def _grab_task(self) -> WorkerTask:
        """Will initiate that a task is gotten from the queue and that it
        spawns its worker process.

        Returns:
            WorkerTask: The WorkerTask grabbed from the queue.

        Raises:
            queue.Empty: If the task queue was empty
        """

        # Get a task from the queue
        try:
            task = self.task_queue.get_nowait()

        except queue.Empty as err:
            raise queue.Empty(
                "No more tasks available in tasks queue."
            ) from err

        else:
            log.debug(
                "Got task %s from queue. (Priority: %s)",
                task.uid,
                task.priority,
            )

        # Let it spawn its own worker
        task.spawn_worker()

        # Now return the task
        return task

    def _assign_tasks(self, num_free_workers: int) -> int:
        """Assigns tasks to (at most) ``num`` free workers."""
        if num_free_workers <= 0:
            return 0

        # How many workers are meant to be spawned?
        if self.spawn_rate == -1:
            num_spawn = num_free_workers
        else:
            num_spawn = min(num_free_workers, self.spawn_rate)

        # Grab the tasks, spawning the corresponding workers
        num_spawned = 0
        for _ in range(num_spawn):
            # Try to grab a task and start working on it
            try:
                new_task = self._grab_task()

            except queue.Empty:
                # There were no tasks left in the task queue
                break

            else:
                # Succeeded in grabbing a task; worker spawned
                self.active_tasks.append(new_task)
                num_spawned += 1

        return num_spawned

    def _poll_workers(self) -> None:
        """Will poll all workers that are in the working list and remove them
        from that list if they are no longer alive.
        """
        # Poll the task's worker's status
        for task in self.active_tasks:
            if task.worker_status is not None:
                # This task has finished. Need to rebuild the list
                break
        else:
            # Nothing to rebuild
            return

        # Broke out of the loop, i.e.: at least one task finished
        old_len = len(self.active_tasks)

        # have to rebuild the list of active tasks now...
        self.active_tasks[:] = [
            t for t in self.active_tasks if t.worker_status is None
        ]
        # NOTE this will also poll all other active tasks and potentially not
        #      add them to the active_tasks list again.

        # Now, only active tasks are in the list, but the list is shorter
        # Can deduce the number of finished tasks from this
        self._num_finished_tasks += old_len - len(self.active_tasks)

        return

    def _check_stop_conds(
        self, stop_conds: Sequence[StopCondition]
    ) -> Set[WorkerTask]:
        """Checks the given stop conditions for the active tasks and compiles
        a list of tasks that needs to be terminated.

        Args:
            stop_conds (Sequence[StopCondition]): The stop conditions that
                are to be checked.

        Returns:
            List[WorkerTask]: The WorkerTasks whose workers need to be
                terminated
        """
        to_terminate = []
        log.debug("Checking %d stop condition(s) ...", len(stop_conds))

        for sc in stop_conds:
            log.debug("Checking stop condition '%s' ...", sc.name)

            # Compile the list of tasks that fulfil a stop condition
            fulfilled = [
                t
                for t in self.active_tasks
                if (t not in self.stopped_tasks and sc.fulfilled(t))
            ]

            if fulfilled:
                log.debug(
                    "Stop condition '%s' fulfilled for %d task(s):  %s",
                    sc.name,
                    len(fulfilled),
                    ", ".join([t.name for t in fulfilled]),
                )
                to_terminate += fulfilled

        # Return as set to be sure that they are unique
        return set(to_terminate)

    def _invoke_periodic_callbacks(self):
        """Invokes the ``periodic`` callback function of each active task."""
        for task in self.active_tasks:
            task._invoke_callback("periodic")

    def _signal_workers(
        self, tasks: Union[str, List[WorkerTask]], *, signal: Union[str, int]
    ) -> None:
        """Send signals to a list of WorkerTasks.

        Args:
            tasks (Union[str, List[WorkerTask]]): strings 'all' or 'active' or
                a list of WorkerTasks to signal
            signal (Union[str, int]): The signal to send
        """
        if isinstance(tasks, str):
            if tasks == "all":
                tasks = self.tasks
            elif tasks == "active":
                tasks = self.active_tasks
            else:
                raise ValueError(
                    f"Tasks cannot be specified by string '{tasks}', "
                    "allowed strings are: 'all', 'active'."
                )

        if not tasks:
            log.debug("No worker tasks to signal.")
            return

        log.debug("Sending signal %s to %d task(s) ...", signal, len(tasks))
        for task in tasks:
            task.signal_worker(signal)

        log.debug(
            "All tasks signalled. Tasks' worker status:\n  %s",
            ", ".join([str(t.worker_status) for t in tasks]),
        )

    def _handle_pending_exceptions(self) -> None:
        """This method handles the list of pending exceptions during working,
        starting from the one added most recently.

        As the WorkerManager occupies the main thread, it is difficult for
        other threads to signal to the WorkerManager that an exception
        occurred.
        The pending_exceptions attribute allows such a handling; child threads
        can just add an exception object to it and they get handled during
        working of the WorkerManager.

        This method handles the following exception types in a specific manner:

            - ``WorkerTaskStopConditionFulfilled``: never raising or logging
            - ``WorkerTaskNonZeroExit``: raising or logging depending on the
              value of the ``nonzero_exit_handling`` property

        Returns:
            None

        Raises:
            Exception: The exception that was added first to the queue of
                pending exceptions
        """

        def log_task_stream(
            task: WorkerTask, *, num_entries: int, stream_name: str = "out"
        ) -> None:
            """Logs the last `num_entries` from the log of the `stream_name`
            of the given WorkerTask object using log.error
            """
            # Get the stream
            stream = task.streams[stream_name]

            # Get lines and print stream using logging module
            lines = stream["log"][-num_entries:]
            log.error(
                "Last ≤%d lines of combined stdout and stderr:\n" "\n  %s\n",
                num_entries,
                "\n  ".join(lines),
            )

        if self.pending_exceptions.empty():
            log.debug("No exceptions pending.")
            return
        # else: there was at least one exception

        # Go over all exceptions
        while not self.pending_exceptions.empty():
            # Get one exception off the queue
            exc = self.pending_exceptions.get_nowait()

            # Currently, only WorkerTaskNonZeroExit exceptions are handled here
            # If the type does not match, can directly raise it
            if not isinstance(exc, WorkerTaskNonZeroExit):
                log.error(
                    "Encountered a pending exception that requires " "raising!"
                )
                raise exc

            log.debug("Handling %s ...", exc.__class__.__name__)

            # Take care of stop conditions, which do NOT need to be raised
            if isinstance(exc, WorkerTaskStopConditionFulfilled):
                continue

            # Distinguish different ways of handling these exceptions
            # Ignore all
            if self.nonzero_exit_handling == "ignore":
                # Done here. Continue with the next exception
                continue

            # Ignore terminated tasks for `warn` and `ignore` levels
            elif abs(exc.task.worker_status) == SIGMAP[
                "SIGTERM"
            ] and self.nonzero_exit_handling not in ["warn_all", "raise"]:
                continue

            # else: will generate some log output, so need to adjust Reporter
            #       to not use CR characters which would mangle up the output
            if self.reporter:
                self.reporter.suppress_cr = True

            # Provide some info on the exit status

            if self.nonzero_exit_handling in ["warn", "warn_all"]:
                # Print the error and the last few lines of the error log
                log.warning(str(exc))
                log_task_stream(exc.task, num_entries=8)

                # Nothing else to do
                continue

            # At this stage, the handling mode is 'raise'. Show more log lines:
            log.critical(str(exc))
            log_task_stream(exc.task, num_entries=24)

            # By raising here, the except block in start_working will be
            # invoked and terminate workers before calling sys.exit
            raise exc

        # The pending_exceptions list is now empty
        log.debug("Handled all pending exceptions.")
